from ..base import ArduinoDevice
from raspidevkit.constants import INPUT
from typing import Tuple, Optional, Union


class HallEffectSensor(ArduinoDevice):
    def __init__(self, arduino, pin: int, commands: dict[str, Union[str, int]]) -> None:
        """
        Create a hall effect sensor object attached to Arduino
        :param arduino: Arduino master
        :param pin: Pin to attach hall effect sensor in Arduino
        :param commands: Dictionary of commands with methods as keys.
        Required command configuration:
        ```python
        commands = {
            "read": 1,
        }
        ```
        """
        pin_setup = {
            str(pin): INPUT
        }
        super().__init__(arduino=arduino,
                         pin_setup=pin_setup, 
                         device_type=INPUT, 
                         commands=commands)
        all_methods = [
            'read',
        ]
        self.validate_commands(all_methods)
        self._method_code = self._map_method_code()

        self._state = False
        self._code_mapping['setup'] = f'pinMode({pin}, INPUT);'
        self._code_mapping['methods'] = self._method_code


    def _map_method_code(self):
        """
        Create method code mapping
        """
        read = f'sendResponse(String(digitalRead({self.pin})));'

        method_code = {
            'read': read
        }
        return method_code



    def read(self) -> Tuple[Optional[float], Optional[float]]:
        """
        Get hall effect sensor state
        :return: pressed state
        """
        command = self._commands.get('read')
        self._arduino.send_command(command)
        response = self._arduino.read_response(origin='cmd')
        try:
            if response == "1":
                return True
            return False
        except:
            raise Exception('Hall effect sensor read fail')



    def __str__(self):
        """
        Device name
        """
        return "hall_effect_sensor"



    def __repr__(self):
        return f"Arduino Hall Effect Sensor <pin={self.pin}"



import sys
sys.path.append('./')

from raspidevkit import Machine
from raspidevkit import (
    RED, GREEN, BLUE, YELLOW, MAGENTA, CYAN, WHITE
)
import pytest
import time


@pytest.fixture()
def machine():
    machine = Machine()
    yield machine
    machine.cleanup()


class TestGpioDevice:
    def test_button(self, machine: Machine):
        button = machine.attach_button(18)
        button.read()


    def test_hall_effect_sensor(self, machine: Machine):
        hall_effect_sensor = machine.attach_hall_effect_sensor(18)
        hall_effect_sensor.read()

    
    def test_buzzer_active(self, machine: Machine):
        active_buzzer = machine.attach_active_buzzer(18)
        active_buzzer.turn_on()
        time.sleep(1)
        active_buzzer.turn_off()


    def test_buzzer_passive(self, machine: Machine):
        passive_buzzer = machine.attach_passive_buzzer(18)
        passive_buzzer.buzz(515)
        time.sleep(1)
        passive_buzzer.play_note('E2')

    
    def test_dht_sensors(self, machine: Machine):
        # Cant mock test
        dht11 = machine.attach_dht11(18)
        dht22 = machine.attach_dht22(14)



    def test_led(self, machine: Machine):
        led = machine.attach_led(18)
        led.turn_on()
        time.sleep(1)
        led.turn_off()


    def test_light_sensor(self, machine: Machine):
        light_sensor = machine.attach_light_sensor(18)
        light_sensor.read()


    def test_pir_motion_sensor(self, machine: Machine):
        pir_motor_sensor = machine.attach_pir_motion_sensor(18)
        pir_motor_sensor.read()


    def test_relay(self, machine: Machine):
        relay = machine.attach_relay(18)
        relay.turn_on()
        time.sleep(1)
        relay.turn_off()


    def test_rgb_led(self, machine: Machine):
        rgb_pins = (14, 15, 18)
        colors = [RED, GREEN, BLUE, YELLOW, MAGENTA, CYAN, WHITE]
        rgb_led = machine.attach_rgb_led(rgb_pins)
        rgb_led.turn_on()
        time.sleep(1)
        for color in colors:
            rgb_led.set_color(color)
            time.sleep(1)
        rgb_led.turn_off()


    def test_servo_motor(self, machine: Machine):
        servo_motor = machine.attach_servo_motor(18)
        servo_motor.rotate(90)
        time.sleep(1)
        servo_motor.rotate(0)


    def test_ultrasonic_sensor(self, machine: Machine):
        echo_pin = 15
        trigger_pin = 18
        ultrasonic_sensor = machine.attach_ultrasonic_sensor((echo_pin, trigger_pin))
        ultrasonic_sensor.get_distance()
        ultrasonic_sensor.get_distance(measure='inch')
        ultrasonic_sensor.get_distance(measure='ft')
        ultrasonic_sensor.get_distance(measure='m')
        with pytest.raises(Exception):
            ultrasonic_sensor.get_distance(measure='km')
        


from ..base import ArduinoDevice
from raspidevkit.constants import INPUT
from typing import Tuple, Optional, Union
import time


class DHTSensor(ArduinoDevice):
    def __init__(self, arduino, pin: int, dtype: str, commands: dict[str, Union[str, int]]) -> None:
        """
        Create a LED object attached to Arduino

        :param arduino: Arduino master
        :param pin: Pin to attach DHT sensor in Arduino
        :param dtype: DHT sensor type
        :param commands: Dictionary of commands with methods as keys.
        Required command configuration:
        ```python
        commands = {
            "get_data": 1,
            "get_temperature": 2,
            "get_humidity": 3
        }
        ```
        """
        pin_setup = {
            str(pin): INPUT
        }
        super().__init__(arduino=arduino,
                         pin_setup=pin_setup, 
                         device_type=INPUT, 
                         commands=commands)
        all_methods = [
            'get_data',
            'get_temperature',
            'get_humidity'
        ]
        self.__var_name = f'dht{self.uuid}'
        self.validate_commands(all_methods)
        self._method_code = self._map_method_code()

        self.__type = dtype
        self._state = False
        self._code_mapping['libraries'] = ['DHT.h']
        self._code_mapping['global'] = f'DHT {self.__var_name} = DHT({pin},{dtype});'
        self._code_mapping['setup'] = f'{self.__var_name}.begin();'
        self._code_mapping['methods'] = self._method_code

    
    def _map_method_code(self):
        """
        Create method code mapping
        """
        get_data_code = f'sendResponse(String({self.__var_name}.readTemperature()) + " " + String({self.__var_name}.readHumidity()));'
        get_temperature_code = f'sendResponse(String({self.__var_name}.readTemperature()));'
        get_humidity_code = f'sendResponse(String({self.__var_name}.readHumidity()));'

        method_code = {
            'get_data': get_data_code,
            'get_temperature': get_temperature_code,
            'get_humidity': get_humidity_code
        }
        return method_code



    def get_data(self) -> Tuple[Optional[float], Optional[float]]:
        """
        Get temperature and humidity from the sensor
        :return: temperature, humidity
        """
        command = self._commands.get('get_data')
        self._arduino.send_command(command)
        response = self._arduino.read_response(origin='cmd')
        data = response.split()
        try:
            temperature = float(data[0])
            humidity = float(data[1])
            return temperature, humidity
        except:
            raise Exception('Sensor read fail')



    def get_temperature(self) -> float:
        """
        Get temperature from the sensor
        :return: Temperature
        """
        command = self._commands.get('get_temperature')
        self._arduino.send_command(command)
        response = self._arduino.read_response(origin='cmd')
        try:
            temperature = float(response)
            return temperature
        except:
            raise Exception('Sensor read fail')
        


    def get_humidity(self) -> float:
        """
        Get humidity from the sensor
        :return: Humidity
        """
        command = self._commands.get('get_humidity')
        self._arduino.send_command(command)
        response = self._arduino.read_response(origin='cmd')
        try:
            humidity = float(response)
            return humidity
        except:
            raise Exception('Sensor read fail')
        


    def __str__(self):
        """
        Device name
        """
        return "dht_sensor"
    


    def __repr__(self):
        return f"Arduino DHT Sensor <pin={self.pin}, type={self.__type}>"



class DHT11(DHTSensor):
    def __init__(self, arduino, pin: int, 
                 commands: dict[str, Union[str, int]], uuid: str = ''):
        """
        Create a DHT11 object
        :param arduino: Arduino instance as this device parent
        :param pin: Pin this device is attached to
        :param commands: Dictionary of commands with methods as keys.
        """
        super().__init__(arduino=arduino,
                         pin=pin,
                         dtype='DHT11',
                         commands=commands)




    def __repr__(self):
        return f"DHT11 <pin={self.pin}>"
    


class DHT22(DHTSensor):
    def __init__(self, arduino, pin: int, 
                 commands: dict[str, Union[str, int]], uuid: str = ''):
        """
        Create a DHT22 object
        :param arduino: Arduino instance as this device parent
        :param pin: Pin this device is attached to
        :param commands: Dictionary of commands with methods as keys.
        """
        super().__init__(arduino=arduino,
                         pin=pin,
                         dtype='DHT22',
                         commands=commands)




    def __repr__(self):
        return f"DHT22 <pin={self.pin}>"


import sys
sys.path.append('./')

from raspidevkit import Machine
import pytest


@pytest.fixture()
def machine():
    machine = Machine()
    yield machine
    # No cleanup
    # machine.cleanup()


class TestArduino:
    def test_arduino(self, machine: Machine):
        arduino = machine.attach_arduino(port=None)



import sys
sys.path.append('../')

from ..base import GpioDevice
from raspidevkit.constants import INPUT, PULL_UP
    

class HallEffectSensor(GpioDevice):
    def __init__(self, machine, pin: int):
        """
        Create a hall effect sensor object

        :param machine: Machine instance as this device parent
        :param pin: Pin this device is attached to
        """
        pin_setup = {
            str(pin): PULL_UP
        }
        super().__init__(machine, pin_setup, device_type=INPUT)



    def read(self) -> bool:
        """
        Read the current state of hall effect sensor

        :return: State (on/off)
        """
        return self.gpio_read(self.pin)



    def __repr__(self):
        return f"Hall effect sensor <pin={self.pin}, state={self.read()}>"
    


from .__logger import MachineLogger
from .__interface import (
    ArduinoCliInterface,
    GpioInterface,
    I2cInterface
)
from typing import Union
from .utils import dictutils
from .devices import (
    ActiveBuzzer,
    Arduino,
    Button,
    DHT11,
    DHT22,
    HallEffectSensor,
    Led,
    L293DDriver,
    L298NDriver,
    LightSensor,
    PassiveBuzzer,
    PIRMotionSensor,
    Relay,
    RgbLed,
    ServoMotor,
    Sim808,
    UltrasonicSensor
)
import logging


class Machine:
    def __init__(self, 
                 gpio_mode: str = 'BCM',
                 i2cbus: Union[None, int] = None,
                 enable_logging: bool = False,
                 debug: bool = False,
                 arduino_cli_path: Union[None, str] = None,
                 **kwargs) -> None:
        '''
        Create a machine class which different devices can be created from.

        :param gpio_mode: GPIO mode can be `BCM` or `BOARD` as `GPIO.setmode()`
        :param i2cbus: I2C bus to use
        :param enable_logging: Enable logging, configuration can be set by passing in a config dictionary
        :param debug: Set debug to true
        :param arduino_cli_path: Path to Arduino CLI
        :param config: Config dictionary which will take priority. See sample config.

        ```python
        default_config = {
            'logging': {
                'format': '%(asctime)s [%(levelname)s] - %(message)s.',
                'file': 'machine.log',
                'level': logging.INFO if not debug else logging.DEBUG
            }
        }
        ```
        '''
        custom_config = kwargs.get('config', {})
        self.logger = None
        self.__intialize_logger(enable_logging, debug, custom_config)

        self.__gpio = GpioInterface(gpio_mode, self.logger)
        self.__i2c = I2cInterface(i2cbus, self.logger)
        self.__arduino_cli = ArduinoCliInterface(arduino_cli_path, self.logger)

        self._devices = []
        self._pin_mapping = []



    @property
    def gpio(self):
        """
        GPIO interface
        """
        return self.__gpio
    

    
    @property
    def i2c(self):
        """
        I2C interface
        """
        return self.__i2c



    @property
    def arduino_cli(self):
        """
        Arduino CLI interface
        """
        return self.__arduino_cli
    


    @property
    def devices(self) -> list:
        """
        Attached devices
        """
        return self._devices
    


    def __intialize_logger(self, enabled: bool, debug: bool, config: dict):
        """
        Initialize logger. `debug` variable will take priority
        """
        default_format = '%(asctime)s [%(levelname)s] - %(message)s.'
        default_file = 'machine.log'
        format = dictutils.get(config, 'logging', 'format', default=default_format)
        file = dictutils.get(config, 'logging', 'file', default=default_file)
        if not debug:
            level = dictutils.get(config, 'logging', 'level', default=logging.INFO)
        else:
            level = logging.DEBUG

        logging_format = logging.Formatter(format)
        main_handler = logging.FileHandler(file)
        main_handler.setFormatter(logging_format)
        self.logger = MachineLogger(__name__, enabled=enabled or debug)
        self.logger.addHandler(main_handler)
        self.logger.setLevel(level)
        self.logger.info('Logger initialized.')



    ######################################################
    #                                                    #
    #                  Devices Interface                 #
    #                                                    #
    ######################################################
        

    def _validate_pin(self, pin: Union[int, list[int], tuple[int]]):
        """
        Check if pin/pins is available

        :param pin: Pin to check
        :raises Exception: If pin is already in use
        """
        if isinstance(pin, int):
            if pin in self._pin_mapping:
                raise Exception(f'Pin {pin} already in use.')
            self._pin_mapping.append(pin)
        if isinstance(pin, (list, tuple)):
            for p in pin:
                if p in self._pin_mapping:
                    raise Exception(f'Pin {p} already in use.')
                self._pin_mapping.append(p)



    def cleanup(self):
        """
        Perform cleanup for graceful exit
        """
        self.logger.info('Performing cleanup routines.')
        self.gpio.cleanup()
        for device in self._devices:
            device.cleanup()
        self.logger.info('Cleanup finished.')



    def attach_button(self, pin: int) -> Button:
        """
        Attach a button to this machine

        :param pin: Pin to attach the button to
        """
        self._validate_pin(pin)
        button = Button(self, pin)
        self._devices.append(button)
        self.logger.info(f'Button attached to pin: {pin}')
        return button
    


    def attach_hall_effect_sensor(self, pin: int) -> HallEffectSensor:
        """
        Attach a hall effect sensor to this machine

        :param pin: Pin to attach the button to
        """
        self._validate_pin(pin)
        hall_effect_sensor = HallEffectSensor(self, pin)
        self._devices.append(hall_effect_sensor)
        self.logger.info(f'Hall effect sensor attached to pin: {pin}')
        return hall_effect_sensor
    


    def attach_dht11(self, pin: int) -> DHT11:
        """
        Attach a DHT11 sensor to this machine

        :param pin: Pin to attach the sensor to
        """
        self._validate_pin(pin)
        dht11 = DHT11(self, pin)
        self._devices.append(dht11)
        self.logger.info(f'DHT11 sensor attached to pin: {pin}')
        return dht11
    


    def attach_dht22(self, pin: int) -> DHT22:
        """
        Attach a DHT22 sensor to this machine

        :param pin: Pin to attach the sensor to
        """
        self._validate_pin(pin)
        dht22 = DHT22(self, pin)
        self._devices.append(dht22)
        self.logger.info(f'DHT22 sensor attached to pin: {pin}')
        return dht22
    


    def attach_led(self, pin: int) -> Led:
        """
        Attach a LED to this machine

        :param pin: Pin to attach the LED to
        """
        self._validate_pin(pin)
        led = Led(self, pin)
        self._devices.append(led)
        self.logger.info(f'LED attached to pin: {pin}')
        return led
    


    def attach_rgb_led(self, pins: tuple[int]) -> RgbLed:
        """
        Attach a RGB LED to this machine

        :param pins: Tuple of pins to use, in RGB order (red, green, blue)
        """
        self._validate_pin(pins)
        rgb_led = RgbLed(self, pins)
        self._devices.append(rgb_led)
        self.logger.info(f'RGB LED attached to pins: {pins}')
        return rgb_led
    


    def attach_relay(self, pin: int) -> Relay:
        """
        Attach a relay to this machine

        :param pin: Pin to attach the relay to
        """
        self._validate_pin(pin)
        relay = Relay(self, pin)
        self._devices.append(relay)
        self.logger.info(f'Relay attached to pin: {pin}')
        return relay
    


    def attach_active_buzzer(self, pin: int) -> ActiveBuzzer:
        """
        Attach an active buzzer to this machine

        :param pin: Pin to use
        """
        self._validate_pin(pin)
        active_buzzer = ActiveBuzzer(self, pin)
        self._devices.append(active_buzzer)
        self.logger.info(f'Active buzzer attached to pin: {pin}')
        return active_buzzer
    


    def attach_passive_buzzer(self, pin: int) -> PassiveBuzzer:
        """
        Attach a passive buzzer to this machine

        :param pin: Pin to use
        """
        self._validate_pin(pin)
        passive_buzzer = PassiveBuzzer(self, pin)
        self._devices.append(passive_buzzer)
        self.logger.info(f'Passive buzzer attached to pin: {pin}')
        return passive_buzzer
    


    def attach_light_sensor(self, pin: int) -> LightSensor:
        """
        Attach a light sensor to this machine
        A capacitor is needed in the circuit
        Refer to this [diagram](https://pimylifeup.com/wp-content/uploads/2016/01/Light-Sensor-Circuit.jpg)

        :param pin: Pin to use
        """
        self._validate_pin(pin)
        light_sensor = LightSensor(self, pin)
        self._devices.append(light_sensor)
        self.logger.info(f'Light sensor attached to pin: {pin}')
        return light_sensor
    


    def attach_pir_motion_sensor(self, pin: int) -> PIRMotionSensor:
        """
        Attach a PIR motion sensor to this machine

        :param pin: Pin to use
        """
        self._validate_pin(pin)
        pir_motion_sensor = PIRMotionSensor(self, pin)
        self._devices.append(pir_motion_sensor)
        self.logger.info(f'PIR motion sensor attached to pin: {pin}')
        return pir_motion_sensor
    


    def attach_servo_motor(self, pin: int) -> ServoMotor:
        """
        Attach a servo motor to this machine

        :param pin: Pin to use
        """
        self._validate_pin(pin)
        servo_motor = ServoMotor(self, pin)
        self._devices.append(servo_motor)
        self.logger.info(f'Servo motor attached to pin: {pin}')
        return servo_motor
    


    def attach_ultrasonic_sensor(self, pins: tuple[int]) -> UltrasonicSensor:
        """
        Attach a ultrasonic sensor to this machine

        :param pins: Tuple of pins to use, in order (trigger, echo)
        """
        self._validate_pin(pins)
        ultrasonic_sensor = UltrasonicSensor(self, pins)
        self._devices.append(ultrasonic_sensor)
        self.logger.info(f'Ultrasonic sensors attached to pin: {pins}')
        return ultrasonic_sensor
    


    def attach_l298n(self, pins: Union[list, tuple]) -> L298NDriver:
        """
        Attach a L298N Driver to this machine

        :param pins: Pins this device is connected to.
                     Should follow the format 
                     (ena, in1, in2) or (ena, in1, in2, enb, in3, in4)
        """
        self._validate_pin(pins)
        l298n = L298NDriver(self, pins)
        self._devices.append(l298n)
        self.logger.info(f'L298N Driver attached to pins: {pins}')
        return l298n



    def attach_l293d(self, pins: Union[list, tuple]) -> L293DDriver:
        """
        Attach a L293D Driver to this machine

        :param pins: Pins this device is connected to.
                     Should follow the format 
                     (ena, in1, in2) or (ena, in1, in2, enb, in3, in4)
        """
        self._validate_pin(pins)
        l293d = L293DDriver(self, pins)
        self._devices.append(l293d)
        self.logger.info(f'L293D Driver attached to pins: {pins}')
        return l293d
    


    def attach_sim808(self, port: str) -> Sim808:
        """
        Attach a SIM808 module to this machine

        :param port: Port to use
        """
        sim808 = Sim808(port)
        self._devices.append(sim808)
        self.logger.info(f'SIM808 attached to port: {port}')
        return sim808
    

    
    def attach_arduino(self, port: str, **kwargs) -> Arduino:
        """
        Attach an Arduino to this machine

        :param port: Port to use
        """
        arduino = Arduino(self, port, **kwargs)
        self._devices.append(arduino)
        self.logger.info(f'Arduino attached to port: {port}')
        return arduino



import sys
sys.path.append('../')

from ..base import GpioDevice
from raspidevkit.constants import OUTPUT, RED, GREEN, BLUE, YELLOW, MAGENTA, CYAN, WHITE


class RgbLed(GpioDevice):
    def __init__(self, machine, pins: tuple[int, str]) -> None:
        """
        Create a RGB LED device that can change color light

        :param machine: Machine instance as this device parent
        :param pins: Tuple of pins to use, in RGB order (red, green, blue)
        """
        if len(pins) != 3:
            raise ValueError("Pins should have 3 values.")
        
        pin_setup = {
            str(pins[0]): OUTPUT,
            str(pins[1]): OUTPUT,
            str(pins[2]): OUTPUT,
        }
        super().__init__(machine, pin_setup, device_type=OUTPUT)
        self._color = None
        self._state = False



    @property
    def state(self):
        """
        Fevice current state
        """
        return self._state
    


    @property
    def color(self):
        """
        Device current color
        """
        return self._color
    


    def turn_on(self):
        """
        Turn RGB LED on (white). Has no effect if already turned on.
        """
        if not self._state:
            for pin in self.pins:
                self.gpio_write(pin, True)
            self._state = True
            self._color = WHITE



    def turn_off(self):
        """
        Turn RGB LED off. Has no effect if already turned off.
        """
        if self._state:
            for pin in self.pins:
                self.gpio_write(pin, False)
            self._state = False
            self._color = None
            


    def set_color(self, color: str):
        """
        Change RGB LED color

        :param color: Color to set this RGB LED to
        Available colors:
        * RED
        * GREEN
        * BLUE
        * YELLOW
        * MAGENTA
        * CYAN
        * WHITE
        """
        color = color.upper()
        supported_colors = [RED, GREEN, BLUE, YELLOW, MAGENTA, CYAN, WHITE]
        if color not in supported_colors:
            raise NotImplementedError(f'Color {color} is not supported.')

        self._state = True
        if color == RED:
            self.gpio_write(self.pins[0], True)
            self.gpio_write(self.pins[1], False)
            self.gpio_write(self.pins[2], False)
        if color == GREEN:
            self.gpio_write(self.pins[0], False)
            self.gpio_write(self.pins[1], True)
            self.gpio_write(self.pins[2], False)
        if color == BLUE:
            self.gpio_write(self.pins[0], False)
            self.gpio_write(self.pins[1], False)
            self.gpio_write(self.pins[2], True)
        if color == YELLOW:
            self.gpio_write(self.pins[0], True)
            self.gpio_write(self.pins[1], True)
            self.gpio_write(self.pins[2], False)
        if color == MAGENTA:
            self.gpio_write(self.pins[0], True)
            self.gpio_write(self.pins[1], False)
            self.gpio_write(self.pins[2], True)
        if color == CYAN:
            self.gpio_write(self.pins[0], False)
            self.gpio_write(self.pins[1], True)
            self.gpio_write(self.pins[2], True)
        if color == WHITE:
            self.gpio_write(self.pins[0], True)
            self.gpio_write(self.pins[1], True)
            self.gpio_write(self.pins[2], True) 



    def cleanup(self):
        """
        Perform cleanup
        """
        self.turn_off()
        


    def __repr__(self):
        return f"RGB LED <pins={self.pins}, state={self._state}>"
    


from ..constants import INPUT, OUTPUT, PULL_UP, PULL_DOWN
from ..utils import stringutil

from typing import Union
import serial
import sys

try:
    import RPi.GPIO as GPIO
except (RuntimeError, ModuleNotFoundError):
    import fake_rpi
    sys.modules['RPi'] = fake_rpi.RPi
    sys.modules['RPi.GPIO'] = fake_rpi.RPi.GPIO
    import RPi.GPIO as GPIO


class GpioDevice:
    def __init__(self, machine, pin_setup: dict, device_type: str) -> None:
        """
        Create a GPIO device. Pin setup can include both input and output.

        :param machine: Machine instance
        :param pin_setup: Pin setup configuration with keys as pins and values as mode
        :param mode: Pin setup mode, can be `INPUT`, `PULL_UP`, `PULL_DOWN`, or `OUTPUT`
        :param device_type: This device device_type `INPUT` or `OUTPUT`

        Example config:
        ```python
        "setup": {
            "14": INPUT,
            "15": PULL_UP,
            "17": PULL_DOWN,
            "24": OUTPUT
        }
        ```
        """
        from raspidevkit import Machine

        self.__machine: Machine = machine
        for pin, setup in pin_setup.items():
            self.__machine.gpio.setup(int(pin), setup)
        
        self.__multi_pin = False
        if len(pin_setup.keys()) > 1:
            self.__multi_pin = True
        
        if self.__multi_pin:
            self.__pins = tuple([int(pin) for pin in pin_setup.keys()])
        else:
            self.__pin = int(list(pin_setup.keys())[0])
        self._device_type = device_type


    
    @property
    def device_type(self) -> str:
        """
        This device type
        """
        return self._device_type



    @property
    def pin(self) -> int:
        """
        Device pin, if single pin
        """
        if self.__multi_pin:
            raise Exception('This is not accessible for multi-pin device. Use pins instead')
        return self.__pin
    


    @property
    def pins(self) -> tuple[int]:
        """
        Device pins, if multi-pin
        """
        if not self.__multi_pin:
            raise Exception('This is not accessible for single-pin device. Use pin instead')
        return self.__pins



    def gpio_setup(self, pin: int,  setup: str):
        """
        Explicitly setup a GPIO pin

        :param pin: Pin to setup
        :param setup: Pin mode (INPUT or OUTPUT)
        """
        if not self.__multi_pin and pin != self.pin:
            raise ValueError('Pin not mapped to device')
        
        if self.__multi_pin and pin not in self.__pins:
            raise ValueError('Pin not mapped to device')
        
        return self.__machine.gpio.setup(pin, setup)



    def gpio_write(self, pin: int, value: bool):
        """
        Perform a GPIO output on pin

        :param pin: The pin to write
        :param value: Can be HIGH or LOW
        :raises ValueError: If pin is not the device pin
        """
        if not self.__multi_pin and pin != self.pin:
            raise ValueError('Pin not mapped to device')
        
        if self.__multi_pin and pin not in self.__pins:
            raise ValueError('Pin not mapped to device')
        
        return self.__machine.gpio.write(pin, value)



    def gpio_read(self, pin: int) -> bool:
        """
        Read the value from a GPIO pin
        
        :param pin: The pin to read
        :return: Value, either True of False
        """
        if not self.__multi_pin and pin != self.pin:
            raise ValueError('Pin not mapped to device')
        
        if self.__multi_pin and pin not in self.__pins:
            raise ValueError('Pin not mapped to device')
        
        return self.__machine.gpio.read(pin)



    def cleanup(self):
        """
        Perform cleanup
        """
        pass



class PwmDevice(GPIO.PWM):
    def __init__(self, machine, pin: int, frequency: float) -> None:
        """
        Create a PWM device

        :param machine: Machine instance
        :param pin: Pin this device is attached to
        :param frequency: Device frequency
        """
        from raspidevkit import Machine

        super().__init__(pin, frequency)
        self.__pin = pin
        self.__frequency = frequency
        self.__duty_cycle = None
        self.__multi_pin = False
        self._device_type = OUTPUT
        self._state = False
        self.__machine: Machine = machine

    

    @property
    def device_type(self) -> str:
        """
        This device type
        """
        return self._device_type
    

    
    @property
    def pin(self) -> int:
        """
        Device pin, if single pin
        """
        if self.__multi_pin:
            raise Exception('This is not accessible for multi-pin device. Use pins instead')
        return self.__pin
    


    @property
    def frequency(self) -> int:
        """
        Device frequency
        """
        return self.__frequency
    


    @frequency.setter
    def frequency(self, frequency: int):
        self.change_frequency(frequency)



    @property
    def duty_cycle(self) -> float:
        """
        Device duty cycle
        """
        return self.__duty_cycle
    


    @duty_cycle.setter
    def duty_cycle(self, duty_cycle: float):
        self.ChangeDutyCycle(duty_cycle)
        self.__duty_cycle = duty_cycle



    def start(self, duty_cycle: float):
        """
        Start device

        :param duty_cycle: Duty cycle
        """
        if not self._state:
            super().start(duty_cycle)
            self.__duty_cycle = duty_cycle
            self._state = True



    def stop(self):
        """
        Stop device
        """
        if self._state:
            super().stop()
            self.__duty_cycle = None
            self._state = False



    def change_frequency(self, frequency: int):
        """
        Change device frequency

        :param frequency: New frequency
        """
        if self._state:
            self.ChangeFrequency(frequency)
            self.__frequency = frequency


        
    def cleanup(self):
        """
        Perform cleanup
        """
        pass
    


class I2CDevice:
    def __init__(self, machine, address: int) -> None:
        """
        Create an I2C device

        :param machine: Machine instance
        :param address: Device I2C address
        """
        from raspidevkit import Machine
        
        self.__address = address
        self.__machine: Machine = machine


    
    @property
    def address(self) -> int:
        """
        Device I2C address
        """
        return self.__address
    

    
    def read_byte(self, force: bool = False) -> int:
        """
        Read a single byte from an I2C address.

        :param force: Force read flag
        :return: Read byte value
        """
        return self.__machine.i2c.read_byte(self.__address, force)
    


    def read_word_data(self, register: int, force: bool = False) -> int:
        """
        Read a single word (2 bytes) from a given register

        :param register: Register address
        :param force: Force read flag
        :return: 2-byte word
        """
        return self.__machine.i2c.read_word_data(self.__address, register, force)
    


    def read_byte_data(self, register: int, force: bool = False) -> int:
        """
        Read a single byte from a designated register
        
        :param register: Register address
        :param force: Force read flag
        :return: Read byte value
        """
        return self.__machine.i2c.read_byte_data(self.__address, register, force)
    


    def read_block_data(self, register: int, length: int = 32, 
                        force: bool = False) -> list[int]:
        """
        Read a block of data from a given register

        :param register: Register address
        :param length: Block length, default to 32 bytes
        :param force: Force read flag
        :return: List of bytes
        """
        return self.__machine.i2c.read_block_data(self.__address, register, length, force)
    


    def write_byte(self, value: int, force: bool = False):
        """
        Write a single byte to an I2C address

        :param value: Value to write
        :param force: Force write flag
        """
        self.__machine.i2c.write_byte(self.__address, value, force)



    def write_word_data(self, register: int, value:int, force: bool = False):
        """
        Write a single word (2 bytes) to a given register

        :param register: Register address
        :param value: Word value to write
        :param force: Force write flag
        """
        self.__machine.i2c.write_word_data(register, value, force)


        
    def write_byte_data(self, register: int, value: int, force: bool = False):
        """
        Write a byte to a given register

        :param register: Register address
        :param value: Byte value to write
        :param force: Force write flag
        """
        self.__machine.i2c_write_byte_data(self.__address, register, value, force)


    
    def write_block_data(self, register: int, data: list[int], 
                         force: bool = False):
        """
        Write a block of byte data to a given register

        :param register: Register address
        :param data: List of byte values to write
        :param force: Force write flag
        """
        self.__machine.i2c.write_block_data(self.__address, register, data, force)



    def cleanup(self):
        """
        Perform cleanup
        """
        pass



class SerialDevice(serial.Serial):
    def __init__(self, machine, port: Union[str, None] = None, 
                 baudrate: int = 9600, **kwargs) -> None:
        """
        Create a serial device

        :param machine: Machine instancce
        :param port: Serial port to use
        :param baudrate: Serial communication baudrate
        :param **kwargs: Other args for serial.Serial object
        """
        from raspidevkit import Machine

        super().__init__(port, baudrate, **kwargs)
        self.__machine: Machine = machine


    
    def write(self, data: Union[str, bytes], encoding: str = 'utf-8') -> int | None:
        """
        Output the given byte string over the serial port.
        String data is automatically encoded

        :param data: String data to send
        """
        if isinstance(data, str):
            data = bytes(data, encoding)
        self.__machine.logger.debug(f'[SERIAL][WRITE] Attempting write on {self.port}: {data}')
        output =  super().write(data)
        self.__machine.logger.debug(f'[SERIAL][WRITE] Returned: {output}')
        return output



    def read(self, size: int = 1, 
             encoding: Union[str, None] = None) -> Union[bytes, str]:
        """
        Read size bytes from the serial port. 
        If a timeout is set it may return less characters as requested. 
        With no timeout it will block until the requested number of bytes is read.

        :param size: Byte size to read
        :param encoding: If given, decodes output to given encoding format
        """
        self.__machine.logger.debug(f'[SERIAL][READ] Attempting read on {self.port}:{size}')
        output =  super().read(size)
        if encoding:
            output = output.decode(encoding)
        self.__machine.logger.debug(f'[SERIAL][READ] Returned: {output}')
        return output
    


    def read_until(self, expected: bytes = b"\n", 
                   size: Union[int, None] = None, 
                   encoding: Union[str, None] = None) -> bytes:
        """
        Read until an expected sequence is found (' ' by default), the size
        is exceeded or until timeout occurs.

        :param expected: Sequence to end reading
        :param size: Byte size to read
        :param encoding: If given, decodes output to given encoding format
        """
        self.__machine.logger.debug(f'[SERIAL][READ] Attempting read on {self.port}:{expected}:{size}')
        output =  super().read_until(expected, size)
        if encoding:
            output = output.decode(encoding)
        self.__machine.logger.debug(f'[SERIAL][READ] Returned: {output}')
        return output
    


    def read_all(self, encoding: Union[str, None] = None) -> Union[bytes, None]:
        """
        Read all bytes currently available in the buffer of the OS.

        :param encoding: If given, decodes output to given encoding format
        """
        self.__machine.logger.debug(f'[SERIAL][READ] Attempting read on {self.port}:all')
        output =  super().read_all()
        if encoding:
            output = output.decode(encoding)
        self.__machine.logger.debug(f'[SERIAL][READ] Returned: {output}')
        return output



class ArduinoDevice:
    def __init__(self, arduino, pin_setup: dict, device_type: str, 
                 commands: dict[str, Union[str, int]], uuid: str = '') -> None:
        """
        Create a Arduino device.

        :param arduino: Arduino master
        :param pin_setup: Config dictionary of pin mode, can be `INPUT`,`PULL_UP`, 
            \t`PULL_DOWN`, `OUTPUT` or `custom`. When custom, the setup method 
            \tcan be overriden
        :param device_type: This device device_type `INPUT`or `OUTPUT`
        :param commands: Dictionary of commands with methods as keys.
        :param uuid: Optional UUID, if not given would randomly generate
        Sample command configuration:
        ```python
        commands = {
            "turnOn": 1,
            "turnOff": 2
        }
        ```
        """
        from .serial.arduino import Arduino

        self._arduino: Arduino = arduino
        self._setup_code = ""
        for pin in pin_setup.keys():
            if pin_setup[pin] == INPUT:
                self._setup_code += f'pinMode({pin}, INPUT);\n'
            elif pin_setup[pin] == PULL_UP:
                self._setup_code += f'pinMode({pin}, INPUT_PULLUP);\n'
            elif pin_setup[pin] == PULL_DOWN:
                self._setup_code += f'pinMode({pin}, INPUT_PULLDOWN);\n'
            elif pin_setup[pin] == OUTPUT:
                self._setup_code += f'pinMode({pin}, OUTPUT);\n'
            elif pin_setup[pin] == 'custom':
                pass
            else:
                raise ValueError(f'Invalid value for pin {pin}')
            
        if uuid:
            self.__uuid = uuid
        else:
            self.__uuid = stringutil.generate_string(8)

        self.__multi_pin = False
        if len(pin_setup.keys()) > 1:
            self.__multi_pin = True
        
        if self.__multi_pin:
            self.__pins = tuple([int(pin) for pin in dict.keys()])
        else:
            self.__pin = int(list(pin_setup.keys())[0])

        self._device_type = device_type
        self._methods = list(commands.keys())
        self._commands = commands
        self._code_mapping = {
            'libraries' : [],
            'global': '',
            'setup': self._setup_code,
            'methods': {}
        }



    @property
    def uuid(self) -> str:
        """
        Device UUID
        """
        return self.__uuid
    

    
    @property
    def device_type(self) -> str:
        """
        This device type
        """
        return self._device_type



    @property
    def pin(self) -> int:
        """
        Device pin, if single pin
        """
        if self.__multi_pin:
            raise Exception('This is not accessible for multi-pin device. Use pins instead')
        return self.__pin
    


    @property
    def pins(self) -> tuple[int]:
        """
        Device pins, if multi-pin
        """
        if not self.__multi_pin:
            raise Exception('This is not accessible for single-pin device. Use pin instead')
        return self.__pins



    @property
    def code(self) -> dict:
        """
        Code mapping
        """
        return self._code_mapping
    


    @property
    def commands(self) -> dict[str, int]:
        """
        Commands available to this device
        """
        return self._commands
    


    def validate_commands(self, keys: list):
        """
        Validate the commands parameter pass. All required
        key should be present.
        """
        for key in keys:
            if key not in self._commands.keys():
                raise Exception(f'Key {key} is not present in commands')



    def _map_method_code(self):
        """
        Create method code mapping
        """
        pass



    def cleanup(self):
        """
        Perform cleanup
        """
        pass



"""
CONSTANT Utilities
"""

# Hardware Related Constants
BCM = 'BCM'
BOARD = 'BOARD'
RISING = 'RISING'
FALLING = 'FALLING'
BOTH = 'BOTH'
INPUT = 'INPUT'
OUTPUT = 'OUTPUT'
PULL_UP = 'PULL_UP'
PULL_DOWN = 'PULL_DOWN'
HIGH = 'HIGH'
LOW = 'LOW'

# Color Related Constants
RED = 'RED'
GREEN = 'GREEN'
BLUE = 'BLUE'
YELLOW = 'YELLOW'
MAGENTA = 'MAGENTA'
CYAN = 'CYAN'
WHITE = 'WHITE'

# Motor Direction Related Constants
CLOCKWISE = 'CW'
COUNTER_CLOCKWISE = 'CCW'



import serial
import time
import datetime
import re


class Sim808(serial.Serial):
    def __init__(self, machine, port: str | None = None, baudrate: int = 9600, **kwargs) -> None:
        """
        Initialize SIM808 device. 

        :param machine: Machine instance as this device parent
        :param port: Port the SIM808 attach to
        :param baudrate: Baudrate of serial commnication
        :param **kwargs: Other args for serial.Serial object
        """
        super().__init__(port, baudrate, **kwargs)
        self.__initialize()
        self._machine = machine



    def __initialize(self):
        """
        Check if SIM808 exists and functioning
        """
        self.send_command('AT\r\n')
        response = self.read_response()
        if('OK' not in response):
            raise Exception('Error starting sim808')
        self.send_command('AT+CMGF=1\r\n')
        self.read_response()
 


    def cleanup(self):
        """
        Perform cleanup
        """
        self.reset_input_buffer()
        self.reset_output_buffer()
        self.close()



    def read_response(self):
        """
        Get the response from SIM808 Serial COM

        :return: SIM808 response
        """
        response = ''
        while(self.in_waiting):
            bit = self.read()
            response = response + bit.decode()
        return response
 


    def send_command(self, command: str, delay: float = 1):
        """
        Send a command to SIM808 Serial COM

        :param command: Command to send
        :param delay: Delay allows module to recieve command in full
        """
        self.write(command.encode())
        time.sleep(delay)
 


    def send_sms(self, number: str, message: str):
        """
        Send a SMS message

        :param number: Number to send message to. Should contain country code
        :param message: Message to send
        """
        self.send_command('AT+CMGS="' + number + '"\r')
        time.sleep(0.1)
        self.send_command(message + '\x1A\n')
        time.sleep(0.1*(len(message)/5))
        response = self.read_response()
        while 'OK' not in response:
            time.sleep(0.1)
            response = self.read_response()



    def read_unread_sms(self):
        """
        Get unread sms

        :return: unread sms
        """
        self.send_command('AT+CMGL=\"REC UNREAD\"\r\n')
        time.sleep(1)
        response = self.read_response()
        return response



    def get_time(self):
        """
        Get network date and time

        :return: Network date and time
        """
        self.send_sms("+639155882825", "CLOCK COMMAND")
        start_time = datetime.datetime.now()
        time.sleep(3)
        messages = self.read_unread_sms()
        end_time = datetime.datetime.now()
        datetime_pattern = r'\d{2}/\d{2}/\d{2},\d{2}:\d{2}:\d{2}\+32'
        match = re.findall(datetime_pattern, messages)
        if match:
            returned_datetime = datetime.datetime.strptime(match[-1].replace('+32',''),'%y/%m/%d,%H:%M:%S')
            returned_datetime += (end_time - start_time)
        return returned_datetime
    


    def delete_all_sms(self):
        """
        Delete all stored sms (inbox and sent)
        """
        self.send_command('AT+CMGD=1,4\r\n')
        time.sleep(5)



import logging
from logging import Logger, Handler


class MachineLogger(Logger):
    def __init__(self, name: str, enabled: bool) -> None:
        super().__init__(name)
        self.enabled = enabled



    def debug(self, msg, *args, **kwargs):
        if self.isEnabledFor(logging.DEBUG) and self.enabled:
            super().debug(msg, *args, **kwargs)
            if kwargs.get('to_stdout', False):
                print(msg)



    def info(self, msg, *args, **kwargs):
        if self.isEnabledFor(logging.INFO) and self.enabled:
            super().info(msg, *args, **kwargs)
            if kwargs.get('to_stdout', False):
                print(msg)



    def warn(self, msg, *args, **kwargs):
        if self.isEnabledFor(logging.WARN) and self.enabled:
            super().warn(msg, *args, **kwargs)
            if kwargs.get('to_stdout', False):
                print(msg)



    def error(self, msg, *args, **kwargs):
        if self.isEnabledFor(logging.ERROR) and self.enabled:
            super().error(msg, *args, **kwargs)
            if kwargs.get('to_stdout', False):
                print(msg)



    def addHandler(self, hdlr: Handler) -> None:
        if self.enabled:
            return super().addHandler(hdlr)
        


from ..base import SerialDevice, ArduinoDevice
from ..arduino.button import Button
from ..arduino.hall_effect_sensor import HallEffectSensor
from ..arduino.led import Led
from ..arduino.relay import Relay
from ..arduino.servo_motor import ServoMotor
from ..arduino.dht import DHT11, DHT22
from raspidevkit.utils import dictutils, stringutil, fileutil
from typing import Union
from string import Template

import subprocess
import tempfile
import pkg_resources


class Arduino(SerialDevice):
    def __init__(self, 
                 machine, 
                 port: Union[str,None] = None, 
                 baudrate: int = 9600,
                 board: str = 'Arduino Uno',
                 **kwargs) -> None:
        """
        Initialize Arduino device

        :param machine: Machine instance as this device parent
        :param port: Port the Arduino attach to
        :param baudrate: Baudrate of serial commnication
        :param **kwargs: Other args for serial.Serial object
        """
        from raspidevkit import Machine

        super().__init__(machine, port, baudrate, **kwargs)
        self.board = board
        self._cmd_terminator = '\n'
        self._data_terminator = '\r\n'
        self._whitespace_sub = '||'
        self.__commands = [-1]
        self._devices = []
        self._pin_used = []
        self.__machine: Machine = machine



    @property
    def devices(self) -> list:
        """
        Attached devices
        """
        return self._devices
    


    def cleanup(self):
        """
        Perform cleanup
        """
        self.reset_input_buffer()
        self.reset_output_buffer()
        self.close()



    def read_response(self, origin: str) -> str:
        """
        Get the response from Arduino Serial COM. \n
        Uses serial.readline()

        :param origin: Response origin can be (`cmd` or `data`)
        :return: Arduino response
        """
        if origin == 'cmd':
            terminator = self._cmd_terminator
        elif origin == 'data':
            terminator = self._data_terminator
        else:
            raise ValueError('Unknown arduino response origin')
        
        response = self.read_until(terminator.encode(), encoding='utf-8')
        return response.strip()
    


    def send_command(self, command: int):
        """
        Send a command to arduino.

        :param command: Command to set, usually auto-generated when attaching devices
        """
        while True:
            self.write(str(command) + self._cmd_terminator)
            response = self.read_response(origin='cmd')
            if 'ok' in response.lower():
                break
        


    def send_data(self, data: str):
        """
        Send a data to arduino

        :param data: Data to send
        """
        raw_data = f'{data}{self._data_terminator}'.replace(' ', self._whitespace_sub)
        while True:
            self.write(raw_data)
            response = self.read_response(origin='data')
            if 'ok' in response.lower():
                break

        

    def _validate_pin(self, pin: Union[int, list[int]]):
        """
        Check if pin/pins is available

        :param pin: Pin to check
        :raises Exception: If pin is already in use
        """
        if isinstance(pin, int):
            if pin in self._pin_used:
                raise Exception(f'Pin {pin} already in use.')
        if isinstance(pin, list):
            for p in pin:
                if p in self._pin_used:
                    raise Exception(f'Pin {pin} already in use.')
        


    def generate_command_list(self, length: int) -> list:
        """
        Generate a list of available command to use

        :param length: Command count to generate
        :return: Available command list
        """
        last_command = max(self.__commands)
        commands = list(range(last_command + 1, last_command + length + 1))
        self.__commands.extend(commands)
        return commands



    def get_attached_device_type(self, _class: ArduinoDevice) -> int:
        """
        Get the number of instances a device type is attached to this arduino

        :param _class: Arduino Device Class
        :return: Number of instance of device type
        """
        count = 0
        for device in self._devices:
            if device.__class__ == _class:
                count += 1
        return count



    def generate_code(self, output_file: str):
        """
        Generate a code to be uploaded to Arduino
        """
        headers = ''
        global_vars = ''
        setups = ''
        methods = ''
        loop = ''
        template_path = pkg_resources.resource_filename('raspidevkit', 
                                                        'templates/arduino_template.ino')
        for index, device in enumerate(self._devices):
            # Add libraries if any
            libraries = device.code['libraries']
            if libraries:
                for library in libraries:
                    headers += f'#include <{library}>\n'

            # Add global variables
            global_vars += device.code['global']

            # Append setup code
            setups += device.code['setup']

            # Create individual method and register to loop
            device_methods = device.code['methods']
            for method, code in device_methods.items():
                method_name = f'{method}_{str(device)}_{index}'
                method_name = stringutil.snake_to_camel(method_name)
                methods += f'void {method_name}() {{\n {code} }}\n'
        
            # Register to loop
            for method, command in device.commands.items():
                method_name = f'{method}_{str(device)}_{index}'
                method_name = stringutil.snake_to_camel(method_name)
                loop += f'else if(currentCommand == {command}){{ {method_name}(); currentCommand = -1;}}\n\n'

        with open(template_path, 'r') as file:
            template_file = file.read()

        template_file = Template(template_file)
        content = template_file.substitute(header=headers,
                                           global_vars = global_vars,
                                           baudrate=self.baudrate,
                                           cmd_terminator = stringutil.escape_whitespace(self._cmd_terminator),
                                           data_terminator = stringutil.escape_whitespace(self._data_terminator),
                                           whitespace_sub = stringutil.escape_whitespace(self._whitespace_sub),
                                           setup=setups,
                                           loop=loop,
                                           methods=methods)
        
        with open(output_file, 'w') as file:
            file.write(content)

        if self.__machine.arduino_cli.formatting:
            self.__format_code(output_file)



    def compile(self, fresh: bool = False):
        """
        Generates, compile and upload code to the arduino
          NOTE: Devices that are only attached up to this point
                are in included in the sketch.

        :param fresh: Bypass hash checking, will always compile sketch
        """
        name = 'raspidevkit.ino'
        raspidevkit_folder = 'raspidevkit'
        temp_dir = tempfile.gettempdir()
        sketch_path = fr'{temp_dir}/{raspidevkit_folder}'
        if not fileutil.does_path_exists(sketch_path):
            fileutil.create_directory(sketch_path)
        file_name = fr'{sketch_path}/{name}'
        previous_hash = ''
        if fileutil.does_file_exists(file_name):
            previous_hash = fileutil.get_file_hash(file_name)
        self.generate_code(file_name)
        current_hash = fileutil.get_file_hash(file_name)

        if previous_hash == current_hash and not fresh:
            self.__machine.logger.info('Generated code matched with previously generated code. Skipping upload')
            return
        fbqn = self.__machine.arduino_cli.get_boards()[self.board]
        self.close()
        self.__machine.arduino_cli.upload_code(sketch_path, self.port, fbqn)
        self.open()



    def __format_code(self, file):
        """
        Format file to C++ formatting with Clang-format
        """
        format_file = pkg_resources.resource_filename('raspidevkit', 'templates/.clang-format')

        try:
            subprocess.run(['clang-format', '-i', file, 
                            f'-style=file:{format_file}'], 
                           check=True, 
                           shell=True)
        except subprocess.CalledProcessError as e:
            self.__machine.logger.error(f'Error formatting {file}: {e}')
            

            
    def attach_button(self, pin: int) -> Button:
        """
        Attach a button to this arduino

        :param pin: Arduino pin to attach button to
        :return: ArduinoButton
        """
        self._validate_pin(pin)
        self._pin_used.append(pin)
        commands = self.generate_command_list(1)
        methods = ['read']
        command_map = dictutils.map_key_value(methods, commands)
        button = Button(self, pin, command_map)
        self._devices.append(button)
        return button
    


    def attach_hall_effect_sensor(self, pin: int) -> HallEffectSensor:
        """
        Attach a hall effect sensor to this arduino
        :param pin: Arduino pin to attach hall effect sensor to
        :return: ArduinoHallEffectSensor
        """
        self._validate_pin(pin)
        self._pin_used.append(pin)
        commands = self.generate_command_list(1)
        methods = ['read']
        command_map = dictutils.map_key_value(methods, commands)
        hall_effect_sensor = HallEffectSensor(self, pin, command_map)
        self._devices.append(hall_effect_sensor)
        return hall_effect_sensor
      
      
      
    def attach_led(self, pin: int) -> Led:
        """
        Attach a LED to this arduino

        :param pin: Arduino pin to attach LED to
        :return: ArduinoLed
        """
        self._validate_pin(pin)
        self._pin_used.append(pin)
        commands = self.generate_command_list(2)
        methods = ['turn_on', 'turn_off']
        command_map = dictutils.map_key_value(methods, commands)
        led = Led(self, pin, command_map)
        self._devices.append(led)
        return led



    def attach_relay(self, pin: int) -> Relay:
        """
        Attach a relay to this arduino

        :param pin: Arduino pin to attach relay to
        :return: ArduinoRelay
        """
        self._validate_pin(pin)
        self._pin_used.append(pin)
        commands = self.generate_command_list(2)
        methods = ['turn_on', 'turn_off']
        command_map = dictutils.map_key_value(methods, commands)
        relay = Relay(self, pin, command_map)
        self._devices.append(relay)
        return relay
    


    def attach_servo_motor(self, pin: int) -> ServoMotor:
        """
        Attach a servo motor to this arduino

        :param pin: Arduino pin to attach servo motor to
        :return: ServoMotor
        """
        self._validate_pin(pin)
        self._pin_used.append(pin)
        commands = self.generate_command_list(1)
        methods = ['rotate']
        uuid = self.get_attached_device_type(ServoMotor) + 1
        command_map = dictutils.map_key_value(methods, commands)
        servo_motor = ServoMotor(self, pin, command_map, uuid=uuid)
        self._devices.append(servo_motor)
        return servo_motor
    


    def attach_dht11(self, pin: int) -> DHT11:
        """
        Attach a DHT11 sensor to this arduino

        :param pin: Arduino pin to attach DHT11 sensor to
        :return: DHT11
        """
        self._validate_pin(pin)
        self._pin_used.append(pin)
        commands = self.generate_command_list(3)
        methods = ['get_data', 'get_temperature', 'get_humidity']
        uuid = self.get_attached_device_type(DHT11) + 1
        command_map = dictutils.map_key_value(methods, commands)
        dht11 = DHT11(self, pin, command_map, uuid=uuid)
        self._devices.append(dht11)
        return dht11



    def attach_dht22(self, pin: int) -> DHT22:
        """
        Attach a DHT22 sensor to this arduino

        :param pin: Arduino pin to attach DHT22 sensor to
        :return: DHT22
        """
        self._validate_pin(pin)
        self._pin_used.append(pin)
        commands = self.generate_command_list(3)
        methods = ['get_data', 'get_temperature', 'get_humidity']
        uuid = self.get_attached_device_type(DHT22) + 1
        command_map = dictutils.map_key_value(methods, commands)
        dht22 = DHT22(self, pin, command_map, uuid=uuid)
        self._devices.append(dht22)
        return dht22



import sys
sys.path.append('../')

from ..base import GpioDevice
from raspidevkit.utils import mathutil
from raspidevkit.constants import INPUT, OUTPUT, PULL_DOWN
from typing import Tuple, Optional
import time
    

class DHTSensor(GpioDevice):
    def __init__(self, machine, pin: int):
        """
        Create a dht sensor object

        :param machine: Machine instance as this device parent
        :param pin: Pin this device is attached to
        """
        pin_setup = {
            str(pin): PULL_DOWN
        }
        super().__init__(machine, pin_setup, device_type=INPUT)



    def get_data(self) -> Tuple[Optional[float], Optional[float]]:
        """
        Get temperature and humidity from the sensor

        :return: temperature, humidity
        """
        data = []
        tmp = 0

        self.__send_start_signal()
        while not self.gpio_read():
            continue
        while self.gpio_read():
            continue

        while tmp < 40:
            counter = 0
            while not self.gpio_read():
                continue
            while self.gpio_read():
                counter += 1
                if counter > 100:
                    break
            if counter < 8:
                data.append(0)
            else:
                data.append(1)
            tmp += 1

        humidity_bit = data[0:8]
        humidity_point_bit = data[8:16]
        temperature_bit = data[16:24]
        temperature_point_bit = data[24:32]
        checksum_bit = data[32:40]

        humidity = mathutil.bits_to_int(humidity_bit, 2)
        humidity_point = mathutil.bits_to_int(humidity_point_bit, 2)
        temperature = mathutil.bits_to_int(temperature_bit, 2)
        temperature_point = mathutil.bits_to_int(temperature_point_bit, 2)
        checksum = mathutil.bits_to_int(checksum_bit, 2)

        valid = self.__verify_checksum(checksum, humidity, humidity_point, 
                                       temperature, temperature_point)
        
        if valid:
            return float(temperature), float(humidity)
        
        return None, None
    


    def get_temperature(self) -> float:
        """
        Get temperature from the sensor

        :return: Temperature
        """
        temperature, _ = self.get_data()
        return temperature
    


    def get_humidity(self) -> float:
        """
        Get humidity from the sensor

        :return: Humidity
        """
        _, humidity = self.get_data()
        return humidity



    def __send_start_signal(self):
        """
        Send start signal to the sensor
        """
        self.gpio_setup(self.pin, OUTPUT)
        self.gpio_write(self.pin, True)
        time.sleep(0.025)
        self.gpio_write(self.pin, False)
        time.sleep(0.02)
        self.gpio_setup(self.pin, INPUT)



    def __verify_checksum(self, *args) -> bool:
        """
        Verify checksum from data
        """
        checksum = args[0]
        total = sum(args) - checksum
        return checksum == (total & 0xFF)



    def __repr__(self):
        return f"DHT Sensor <pin={self.pin}>"
    


class DHT11(DHTSensor):
    def __init__(self, machine, pin: int):
        """
        Create a DHT11 object

        :param machine: Machine instance as this device parent
        :param pin: Pin this device is attached to
        """
        super().__init__(machine, pin)



    def __repr__(self):
        return f"DHT11 <pin={self.pin}>"



class DHT22(DHTSensor):
    def __init__(self, machine, pin: int):
        """
        Create a DHT22 object

        :param machine: Machine instance as this device parent
        :param pin: Pin this device is attached to
        """
        super().__init__(machine, pin)



    def __repr__(self):
        return f"DHT22 <pin={self.pin}>"
    


from .base import DcMotorDriver, DcMotor
from typing import Union

class L298NDriver(DcMotorDriver):
    def __init__(self, machine, pins: Union[list, tuple]) -> None:
        """
        Create a L298N Driver object. Supports up to two motors.

        :param pins: Pins this device is connected to.
                     Should follow the format 
                     (ena, in1, in2) or (ena, in1, in2, enb, in3, in4)
        """
        if len(pins) != 3 and len(pins) != 6:
            raise ValueError('Incorrect pin format')
        
        self.__motor_pin_setup = {}
        motor_pin_1 = {
            'EN': pins[0],
            'A': pins[1],
            'B': pins[2]
        }
        self.__motor_pin_setup['1'] = motor_pin_1

        motor_pin_2 = None
        if len(pins) == 6:
            motor_pin_2 = {
                'EN': pins[3],
                'A': pins[4],
                'B': pins[5]
            }
            self.__motor_pin_setup['2'] = motor_pin_2

        super().__init__(machine, self.__motor_pin_setup)



    def attach_motor(self, is_pwm: bool = False):
        """
        Attach a motor to this driver

        :param is_pwm: Enable PWM mode, allowing to motor to set speed
        :return: L298NMotor object
        """
        if len(self.motors) >= self._max_motor:
            raise RuntimeError('Driver max capacity reached.')
        
        motor_pins = self.__motor_pin_setup.get(str(len(self.motors) + 1))
        if not motor_pins:
            raise RuntimeError('Missing motor pins.')
        
        motor = L298NMotor(self, motor_pins, is_pwm)
        self._motors.append(motor)
        return motor



    def __repl__(self):
        pins = []
        for motor, pin_setup in self.__motor_pin_setup.items():
            for pin_type, pin in pin_setup.items():
                pins.append(pin)
        return f"L298N DC Driver <pins={pins}, current={len(self._motors)}, max={self._max_motor}>"
    


class L298NMotor(DcMotor):
    def __init__(self, driver, pin_setup: dict[str, str], is_pwm: bool = False) -> None:
        super().__init__(driver, pin_setup, is_pwm)
        self.__pin_setup = pin_setup



    def __repl__(self):
        return f"L298N DC Motor <pin={self.__pin_setup}, pwm_mode={self.is_pwm}, state={self.state}>"
    


import sys
sys.path.append('../')

from ..base import GpioDevice, PwmDevice
from raspidevkit.constants import OUTPUT
from raspidevkit.utils import soundutil
from typing import Union
import time


class ActiveBuzzer(GpioDevice):
    def __init__(self, machine, pin: int):
        """
        Create an active buzzer object

        :param pin: Pin this device is attached to
        """
        pin_setup = {
            str(pin): OUTPUT
        }
        super().__init__(machine, pin_setup, device_type=OUTPUT)
        self._state = False



    @property
    def state(self):
        """
        Device current state
        """
        return self._state
    


    def turn_on(self):
        """
        Turn this buzzer on
        """
        if not self.state:
            self.gpio_write(self.pin, True)
            self._state = True



    def turn_off(self):
        """
        Turn this buzzer off
        """
        if self.state:
            self.gpio_write(self.pin, False)
            self._state = False



class PassiveBuzzer(PwmDevice):
    def __init__(self, machine, pin: int) -> None:
        """
        Create a passive buzzer object

        :param pin: Pin this device is attached to
        """
        super().__init__(machine, pin, 440)
        self._machine = machine



    def buzz(self, frequency: int, duration: float = None):
        """
        Activate buzzer with defined frequency.
        If duration not set, must manually call `stop()`

        :param frequency: Frequency
        :param duration: Play for x seconds (blocking)(optional).
        """
        self.start(50)
        self.change_frequency(frequency)
        if duration:
            time.sleep(duration)
            self.stop()
        


    def play_note(self, note: str, duration: float = 1):
        """
        Play a note
        If duration not set, must manually call `stop()`

        :param note: a note like string (e.g. F4, FS3)
        :param duration: Play for x seconds
        """
        self.start(50)
        frequency = soundutil.get_note_frequency(note)
        self.buzz(frequency, duration)



    def play_music(self, notes: Union[list[str], tuple[str]], beats: Union[list[int], tuple[int]]):
        """
        Play a music from notes and beats

        :param notes: Iterables of note like strings
        :param beats: Iterables of beats
        """
        music = zip(notes, beats)
        for note, beat in music:
            self.play_note(note, beat)



    def __repr__(self):
        return f"Buzzer <pin={self.pin}, state={self._state}>"



import sys
sys.path.append('../')

from ..base import GpioDevice
from raspidevkit.constants import OUTPUT


class Relay(GpioDevice):
    def __init__(self, machine, pin: int):
        """
        Create a Relay object

        :param machine: Machine instance as this device parent
        :param pin: Pin this device is attached to
        """
        pin_setup = {
            str(pin): OUTPUT
        }
        super().__init__(machine, pin_setup, device_type=OUTPUT)
        self._state = False



    @property
    def state(self):
        """
        Device current state
        """
        return self._state
    

    
    def turn_on(self):
        """
        Turn on relay. Has no effect if already turn on.
        """
        if not self._state:
            self.gpio_write(self.pin, True)
            self._state = True



    def turn_off(self):
        """
        Turn off relay. Has no effect if already turn off.
        """
        if self._state:
            self.gpio_write(self.pin, False)
            self._state = False



    def set_state(self, state: bool):
        """
        Turn relay on/off.

        :param state: State to switch to
        """
        if state:
            self.turn_on()
        else:
            self.turn_off()



    def cleanup(self):
        """
        Perform cleanup
        """
        self.turn_off()


        
    def __repr__(self):
        return f"Relay <pin={self.pin}, state={self._state}>"




from ..base import ArduinoDevice
from raspidevkit.constants import OUTPUT
from typing import Union
import time


class ServoMotor(ArduinoDevice):
    def __init__(self, arduino, pin: int, 
                 commands: dict[str, Union[str, int]], uuid: str = '') -> None:
        """
        Create a Servo Motor object attached to Arduino

        :param arduino: Arduino master
        :param pin: Pin to attach Servo Motor in Arduino
        :param commands: Dictionary of commands with methods as keys.
        :param uuid: Optional UUID, if not given would randomly generate
        Required command configuration:
        ```python
        commands = {
            "turn_on": 1,
            "turn_off": 2,
            "blink": 3
        }
        ```
        """
        pin_setup = {
            str(pin): 'custom'
        }
        super().__init__(arduino=arduino,
                         pin_setup=pin_setup, 
                         device_type=OUTPUT, 
                         commands=commands,
                         uuid=uuid)
        all_methods = [
            'rotate'
        ]

        self.validate_commands(all_methods)
        self._method_code = self._map_method_code()

        self._state = False
        self._code_mapping['libraries'] = ['Servo.h']
        self._code_mapping['global'] = f'Servo servo{self.uuid};'
        self._code_mapping['setup'] = f'servo{self.uuid}.attach({pin});servo{self.uuid}.write(0);'
        self._code_mapping['methods'] = self._method_code
    


    def _map_method_code(self):
        """
        Create method code mapping
        """
        rotate_code = f'''String data = recieveData();
        int angle = data.toInt();
        servo{self.uuid}.write(angle);
        '''
        method_code = {
            'rotate': rotate_code,
        }
        return method_code
    


    def rotate(self, angle: int):
        """
        Rotate servo motor to angle

        :param angle: Angle to turn
        """
        if not isinstance(angle, int):
            raise ValueError('Angle should be int')

        command = self._commands.get('turn_on')
        self._arduino.send_command(command)
        self._arduino.send_data(str(angle))



    def cleanup(self):
        """
        Perform cleanup
        """
        self.turn_off()



    def __str__(self):
        """
        Device name
        """
        return "servo_motor"
    


    def __repr__(self):
        return f"Arduino Servo Motor <pin={self.pin}, state={self._state}>"
    


import hashlib
import os
import random
import re
import string
from typing import Union

class dictutils:
    """
    Dictionary utility helper
    """

    @staticmethod
    def get(dictionary: dict, *keys: str, default=None,):
        """
        Traverse a dictionary in a tree-like structure

        :param *keys: Keys to search
        :param default: Value to return if key not found else raise KeyError
        :return: Value
        """
        current_value = dictionary
        for key in keys:
            if isinstance(current_value, dict) and key in current_value:
                current_value = current_value[key]
            else:
                if default:
                    return default
                raise KeyError(f'Key not found: {key}')
        return current_value
    


    @staticmethod
    def map_key_value(keys: list, values: list) -> dict:
        """
        Map a values to keys.

        :param keys: List of keys
        :param values: List of values
        :return: Mapped dictionary
        """
        result_dict = dict(zip(keys, values))
        return result_dict
    


    @staticmethod
    def does_keys_exists(dictionary: dict, keys: list[str], strict: bool = True) -> bool:
        """
        Check if all keys exists in the input dictionary

        :param dictionary: Dictionary to check
        :param keys: List of keys to validate
        :param strict: Should all keys exist in the dictionary
        """

        if strict:
            for key in keys:
                if not key in dictionary.keys():
                    return False
            return True
        
        else:
            for key in keys:
                if key in dictionary.keys():
                    return True
            return False



class stringutil:
    """
    String utility helper
    """

    @staticmethod
    def generate_string(length: int) -> str:
        """
        Randomly generate a string with length

        :param length: Length of string to generate
        :return: Generated string
        """
        res = ''.join(random.choices(string.ascii_uppercase + string.digits, 
                                        k=length))
        return res
    


    @staticmethod
    def snake_to_camel(text: str) -> str:
        """
        Convert a snake_case string to camelCase

        :param text: snake_case string
        :return: camelCase string
        """
        parts = text.split('_')
        camel_case = parts[0] + ''.join(word.capitalize() for word in parts[1:])
        return camel_case
    


    @staticmethod
    def escape_whitespace(text: str) -> str:
        """
        Escape whitespace characters

        :param text: Raw string
        :return: Escaped string
        """
        chars = {
            '\r': '\\r',
            '\n': '\\n',
            '\t': '\\t',
            '\f': '\\f',
            '\v': '\\v'
        }
        result = text
        for key, value in chars.items():
            result = result.replace(key, value)
        return result
    


class fileutil:
    """
    File utility helper
    """

    @staticmethod
    def does_file_exists(file_path: str) -> bool:
        """
        Check if a file exists

        :param file_path: File path to check
        :return: Exists or not
        """
        file_path = os.path.normpath(file_path)
        return os.path.exists(file_path)
    


    @staticmethod
    def does_path_exists(path: str) -> bool:
        """
        Check if a path exists

        :param file_path: Path to check
        :return: Exists or not
        """
        path = os.path.normpath(path)
        return os.path.exists(path)
    


    @staticmethod
    def get_parent_directory(file_path: str) -> str:
        """
        Get the parent directory of a file

        :param file_path: File to process
        :return: Parent directory
        """
        return os.path.dirname(file_path)
    


    @staticmethod
    def create_directory(directory: str, cwd: str = None):
        """
        Create a directory and its parent directory if not exist

        :param directory: Directory path to create
        :param cwd: The current working directory
        """
        prev_cwd = os.getcwd()
        if cwd:
            os.chdir(cwd)
        os.makedirs(directory)
        os.chdir(prev_cwd)



    @staticmethod
    def get_absolute_path(file_path: str) -> str:
        """
        Get the absolute file path of a path

        :param file_path: Relative file path
        :return: Absolute file path
        """
        return os.path.abspath(file_path)



    @staticmethod
    def get_file_hash(file_path: str, algorithm: str='sha256') -> str:
        """
        Calculate and get the file hash of a given file

        :param file_path: File to process
        :param algorithm: Hash algorithm to use
        :return: File hash
        """
        hasher = hashlib.new(algorithm)
        with open(file_path, 'rb') as f:
            for chunk in iter(lambda: f.read(4096), b''):
                hasher.update(chunk)
        return hasher.hexdigest()


    
class soundutil:
    """
    Sound utility helper
    """

    @staticmethod
    def get_note_frequency(note: str) -> int:
        """
        Get equivalent frequency of a note

        :param note: Note
        :return: Frequency
        :raises: Value Error if note is not supported 
        """
        notes = {'B0': 31, 'C1': 33, 'CS1': 35, 'D1': 37, 'DS1': 39, 
                 'E1': 41, 'F1': 44, 'FS1': 46, 'G1': 49, 'GS1': 52, 
                 'A1': 55, 'AS1': 58, 'B1': 62, 'C2': 65, 'CS2': 69, 
                 'D2': 73, 'DS2': 78, 'E2': 82, 'F2': 87, 'FS2': 93, 
                 'G2': 98, 'GS2': 104, 'A2': 110, 'AS2': 117, 'B2': 123, 
                 'C3': 131, 'CS3': 139, 'D3': 147, 'DS3': 156, 'E3': 165, 
                 'F3': 175, 'FS3': 185, 'G3': 196, 'GS3': 208, 'A3': 220, 
                 'AS3': 233, 'B3': 247, 'C4': 262, 'CS4': 277, 'D4': 294, 
                 'DS4': 311, 'E4': 330, 'F4': 349, 'FS4': 370, 'G4': 392, 
                 'GS4': 415, 'A4': 440, 'AS4': 466, 'B4': 494, 'C5': 523, 
                 'CS5': 554, 'D5': 587, 'DS5': 622, 'E5': 659, 'F5': 698, 
                 'FS5': 740, 'G5': 784, 'GS5': 831, 'A5': 880, 'AS5': 932, 
                 'B5': 988, 'C6': 1047, 'CS6': 1109, 'D6': 1175, 'DS6': 1245, 
                 'E6': 1319, 'F6': 1397, 'FS6': 1480, 'G6': 1568, 'GS6': 1661, 
                 'A6': 1760, 'AS6': 1865, 'B6': 1976, 'C7': 2093, 'CS7': 2217, 
                 'D7': 2349, 'DS7': 2489, 'E7': 2637, 'F7': 2794, 'FS7': 2960, 
                 'G7': 3136, 'GS7': 3322, 'A7': 3520, 'AS7': 3729, 'B7': 3951, 
                 'C8': 4186, 'CS8': 4435, 'D8': 4699, 'DS8': 4978
                }
        frequency = notes.get(note.upper())
        if not frequency:
            raise ValueError('Note is not supported.')
        return frequency



class mathutil:
    """
    Math utility helper
    """

    @staticmethod
    def map_range(x: float, in_min: float, in_max: 
                  float, out_min: float, out_max: float) -> float:
        """
        Get the x equivalent based on min and max value.
        Just like `map` function of Arduino

        :param x: Input to check for equivalent
        :param in_min: Input lowest value
        :param in_max: Input highest value
        :param out_min: Output lowest value
        :param out_max: Output lowest value
        :return: Output equivalent value
        """
        return (x - in_min) * (out_max - out_min) // (in_max - in_min) + out_min
    


    @staticmethod
    def convert_distance(distance: float, measure_in: str, measure_out: str) -> float:
        """
        Convert the given distance from `measure_in` to `measure_out`
        Available measures: `cm`, `inch`, `ft`, `m`

        :param distance: Input distance
        :param measure_in: Input measure unit
        :param measure_out: Output measure unit
        :return: Converted distance
        """
        conversion_factors = {
            'cm': {'inch': 0.393701, 'ft': 0.0328084, 'm': 0.01},
            'inch': {'cm': 2.54, 'ft': 0.0833333, 'm': 0.0254},
            'ft': {'cm': 30.48, 'inch': 12, 'm': 0.3048},
            'm': {'cm': 100, 'inch': 39.3701, 'ft': 3.28084}
        }

        if measure_in not in conversion_factors \
            or measure_out not in conversion_factors:
            raise ValueError("Invalid input or output measure unit")

        if not isinstance(distance, (int, float)):
            raise ValueError("Distance must be a number")

        if measure_in == measure_out:
            return distance
        
        factor = dictutils.get(conversion_factors, measure_in, measure_out)
        converted_distance = distance * factor
        return converted_distance



    @staticmethod
    def bits_to_int(bits: list[Union[str, int]], base: int):
        """
        Convert a list of bits to int

        :param bits: List of bits
        :param base: N-base of each bits
        """
        bits_str = ''.join([str(bit) for bit in bits])
        output = int(bits_str, base)
        return output



from .machine import Machine
from .constants import (
    BCM, BOARD, RISING, FALLING, BOTH, 
    INPUT, OUTPUT, PULL_UP, PULL_DOWN, 
    RED, GREEN, BLUE, YELLOW, MAGENTA, CYAN, WHITE,
    CLOCKWISE, COUNTER_CLOCKWISE
)

__all__ = [
    'Machine',
    'constants'
]



import sys
sys.path.append('./')

from raspidevkit import Machine
from raspidevkit import COUNTER_CLOCKWISE, CLOCKWISE
import pytest
import time


@pytest.fixture()
def machine():
    machine = Machine()
    yield machine
    machine.cleanup()


class TestDcMotorDriver:
    def test_l293d_single_motor(self, machine: Machine):
        pins = (14, 15, 18)
        l239d = machine.attach_l293d(pins)
        motor = l239d.attach_motor()
        with pytest.raises(Exception):
            motor.set_speed(50)
        motor.run()
        time.sleep(1)
        motor.set_direction(COUNTER_CLOCKWISE)
        time.sleep(1)
        motor.set_direction(CLOCKWISE)
        time.sleep(1)
        motor.stop()


    def test_l293d_single_pwm_motor(self, machine: Machine):
        pins = (14, 15, 18)
        l239d = machine.attach_l293d(pins)
        motor = l239d.attach_motor(is_pwm=True)
        motor.set_speed(50)
        motor.run()
        time.sleep(1)
        motor.set_direction(COUNTER_CLOCKWISE)
        time.sleep(1)
        motor.set_direction(CLOCKWISE)
        time.sleep(1)
        motor.stop()


    def test_l293d_single_setup_multi_motor(self, machine: Machine):
        pins = (14, 15, 18)
        l239d = machine.attach_l293d(pins)
        l239d.attach_motor()
        with pytest.raises(Exception):
            l239d.attach_motor()


    def test_l293d_dual_setup_multi_motor(self, machine: Machine):
        pins = (14, 15, 18, 25, 24, 23)
        l239d = machine.attach_l293d(pins)
        l239d.attach_motor()
        l239d.attach_motor()
        with pytest.raises(Exception):
            l239d.attach_motor()


    def test_l298n_single_motor(self, machine: Machine):
        pins = (14, 15, 18)
        l298n = machine.attach_l298n(pins)
        motor = l298n.attach_motor()
        with pytest.raises(Exception):
            motor.set_speed(50)
        motor.run()
        time.sleep(1)
        motor.set_direction(COUNTER_CLOCKWISE)
        time.sleep(1)
        motor.set_direction(CLOCKWISE)
        time.sleep(1)
        motor.stop()


    def test_l298n_single_pwm_motor(self, machine: Machine):
        pins = (14, 15, 18)
        l298n = machine.attach_l298n(pins)
        motor = l298n.attach_motor(is_pwm=True)
        motor.set_speed(50)
        motor.run()
        time.sleep(1)
        motor.set_direction(COUNTER_CLOCKWISE)
        time.sleep(1)
        motor.set_direction(CLOCKWISE)
        time.sleep(1)
        motor.stop()


    def test_l298n_single_setup_multi_motor(self, machine: Machine):
        pins = (14, 15, 18)
        l298n = machine.attach_l298n(pins)
        l298n.attach_motor()
        with pytest.raises(Exception):
            l298n.attach_motor()


    def test_l298n_dual_setup_multi_motor(self, machine: Machine):
        pins = (14, 15, 18, 25, 24, 23)
        l298n = machine.attach_l298n(pins)
        l298n.attach_motor()
        l298n.attach_motor()
        with pytest.raises(Exception):
            l298n.attach_motor()



from raspidevkit.constants import OUTPUT, CLOCKWISE, COUNTER_CLOCKWISE
from raspidevkit.utils import dictutils
from ...base import GpioDevice, PwmDevice
from abc import abstractmethod


class DcMotorDriver:
    def __init__(self, machine, motor_pin_setup: dict[str, dict[str, str]]) -> None:
        """
        Base class for DcMotorDrivers object.

        :param motor_pin_setup: Pin connection of GPIO pin and physical driver pin.
                          Pins are only initialized when calling `attach_motors()`.
                          Has currently support only on 3 pin per motor drivers (en, a, b)

        Example config:
        ```python
        motor_pin_setup = {
            "1": {
                "EN": "14",
                "A": "15",
                "B": "16"
            },
            "2": {
                "EN": "17",
                "A": "18",
                "B": "19"
            }
        }
        ```
        """
        from  raspidevkit import Machine

        self.__machine: Machine = machine
        self.__gpio = self.__machine.gpio
        self._max_motor = len(motor_pin_setup.keys())
        self._motors = []



    @property
    def gpio(self):
        """
        GPIO interface
        """
        return self.__gpio


    
    @property
    def motors(self):
        """
        Currently attached motors to this driver
        """
        return self._motors
    


    @abstractmethod
    def attach_motor(self):
        """
        Attach a motor to this driver
        """
        raise NotImplementedError('Must be overriden by derived class')



    def gpio_write(self, pin: int, value: bool):
        """
        Perform a GPIO output on a pin

        :param pin: The pin to write
        :param value: Can be HIGH or LOW
        """
        self.__machine.gpio.write(pin, value)



    def gpio_setup(self, pin: int,  setup: str):
        """
        Explicitly setup a GPIO pin

        :param pin: Pin to setup
        :param setup: Pin mode (INPUT or OUTPUT)
        """
        self.__machine.gpio.setup(pin, setup)



    def cleanup(self):
        """
        Perform cleanup for graceful exit
        """
        for motor in self.motors:
            motor.cleanup()
        


class DcMotor(GpioDevice, PwmDevice):
    def __init__(self, driver: DcMotorDriver, pin_setup: dict[str, str], is_pwm: bool = False) -> None:
        """
        Base class for DcMotors object. Must be spawned by a DcMotorDriver class.

        :param pin_setup: Pin connection of driver pin and GPIO pin.
                          Has currently support only on 3 pin per motor drivers (en, a, b).
        :param is_pwm: Enable PWM mode, allowing to motor to set speed

        Example config:
        ```python
        pin_setup = {
            "EN": "14",
            "A": "15",
            "B": "16"
        }
        ```
        """
        self.__driver = driver
        self.__is_pwm = is_pwm
        self._state = False
        required_keys = ['EN', 'A', 'B']

        if not dictutils.does_keys_exists(pin_setup, required_keys):
            raise ValueError('Incorrect pin setup. Missing keys')
        
        self.__pin_en = pin_setup.get('EN')
        self.__pin_a = pin_setup.get('A')
        self.__pin_b = pin_setup.get('B')

        gpio_pins = {}
        for _, pin in pin_setup.items():
            gpio_pins[pin] = OUTPUT

        super().__init__(machine=self.__driver, pin_setup=gpio_pins, device_type=OUTPUT)

        if is_pwm:
            super(PwmDevice, self).__init__(pin_setup.pop('EN'), frequency=1000)
            self.__is_pwm = True
            
        self.__driver.gpio_write(self.__pin_a, True)
        self.__driver.gpio_write(self.__pin_b, False)

        self.__direction = CLOCKWISE
        self.__speed = 100



    @property
    def is_pwm(self):
        """
        PWM is enabled
        """
        return self.__is_pwm
    


    @property
    def state(self):
        """
        Motor state
        """
        return self._state
    


    @property
    def speed(self):
        """
        Motor speed. 100 is highest, 0 is lowest.
        """
        return self.__speed
    


    @speed.setter
    def speed(self, speed: int):
        self.set_speed(speed)



    @property
    def direction(self):
        """
        Current spinning direction
        """
        return self.__direction
    


    @direction.setter
    def direction(self, direction: str):
        self.set_direction(direction)



    def cleanup(self):
        """
        Perform cleanup for graceful exit
        """
        self.stop()



    def run(self):
        """
        Start the motor
        """
        if self._state:
            return
        
        if self.__is_pwm:
            self.start(self.__speed)
        else:
            self.__driver.gpio_write(self.__pin_en, True)
        self._state = True



    def stop(self):
        """
        Stop the motor
        """
        if not self._state:
            return
        
        if self.__is_pwm:
            super().stop()
        else:
            self.__driver.gpio_write(self.__pin_en, False)
        self._state = False


    
    def set_speed(self, speed: int):
        """
        Set motor speed. Only available if `is_pwm`

        :param speed: New speed
        """
        if not self.is_pwm:
            raise RuntimeError('Only available if in pwm mode')
        
        self.duty_cycle = speed
        self.__speed = speed



    def set_direction(self, direction: str):
        """
        Set current spinning direction

        :param direction: New direction to set to. Can be (CLOCKWISE or COUNTER_CLOCKWISE). Use raspidev.constants
        """
        if self.direction == direction:
            return
        
        if direction != CLOCKWISE and direction != COUNTER_CLOCKWISE:
            raise ValueError('Incorrect direction value. Use raspidev.CLOCKWISE or raspidev.COUNTER_CLOCKWISE')

        if direction == CLOCKWISE:
            self.__driver.gpio_write(self.__pin_a, True)
            self.__driver.gpio_write(self.__pin_b, False)
        
        if direction == COUNTER_CLOCKWISE:
            self.__driver.gpio_write(self.__pin_a, False)
            self.__driver.gpio_write(self.__pin_b, True)



import sys
sys.path.append('../')

from ..base import GpioDevice
from raspidevkit.constants import INPUT, OUTPUT
from raspidevkit.utils import mathutil
import time


class UltrasonicSensor(GpioDevice):
    def __init__(self, machine, pins: tuple[int, str]) -> None:
        """
        Create am ultrasonic sensor device

        :param machine: Machine instance as this device parent
        :param pins: Tuple of pins to use, in RGB order (trigger, echo)
        """
        if len(pins) != 2:
            raise ValueError("Pins should have 2 values.")
        
        pin_setup = {
            str(pins[0]): OUTPUT,
            str(pins[1]): INPUT,
        }
        super().__init__(machine, pin_setup, device_type=INPUT)
        self.__trigger_pin = pins[0]
        self.__echo_pin = pins[1]


    
    def get_distance(self, measure: str = 'cm') -> float:
        """
        Get the distance of an object from the sensor

        :param measure: Unit of measurement (`cm`, `inch`, `ft`, `m`)
        :return: Distance of object
        """
        self.gpio_write(self.__trigger_pin, True)
        time.sleep(0.00001)
        self.gpio_write(self.__trigger_pin, False)

        start = time.time()
        end = time.time()

        while not self.gpio_read(self.__echo_pin):
            start = time.time()

        while self.gpio_read(self.__echo_pin):
            end = time.time()
        
        elapsed = end - start
        distance = (elapsed * 34300) / 2
        converted_distance = mathutil.convert_distance(distance, 'cm', measure)
        return converted_distance
    


    def __repr__(self) -> str:
        return f"Ultrasonic Sensor <pins={self.pins}"
    


import sys
sys.path.append('../')

from ..base import PwmDevice
import time


class ServoMotor(PwmDevice):
    def __init__(self, machine, pin: int, frequency: int = 50) -> None:
        """
        Create a RGB LED device that can change color light

        :param machine: Machine instance as this device parent
        :param pin: Pin this device is attached to
        :param frequency: Device frequency
        """
        super().__init__(machine, pin, frequency)
        self.duty_cycle = 0
        self.__min_angle_duty_cycle = (-180, 2.5)
        self.__max_angle_duty_cycle = (180, 12.5)
        self._machine = machine
        self.__angle = 0



    @property
    def angle(self):
        """
        Current angle
        """
        return self.__angle
    


    def set_angle_duty_cycle(self, min: tuple, max: tuple):
        """
        Set the minimum and maximum angle by duty cycle. \n
        Input should be a tuple in the form of (`angle`, `duty_cycle`)

        :param min: The minimum angle by duty cycle
        :param max: The maximum angle by duty cycle
        """
        self.__min_angle_duty_cycle = min
        self.__max_angle_duty_cycle = max



    def rotate(self, angle: int, wait: float = 1):
        """
        Rotate servo at a given angle

        :param angle: Angle to turn servo to
        :param wait: Seconds to wait for rotation to complete
        """
        min_angle = self.__min_angle_duty_cycle[0]
        min_duty_cycle = self.__min_angle_duty_cycle[1]
        max_angle = self.__max_angle_duty_cycle[0]
        max_duty_cycle = self.__max_angle_duty_cycle[1]

        if angle < min_angle:
            raise ValueError('Angle is lower than allowed minimum angle')
        
        if angle > max_angle:
            raise ValueError('Angle is higher than the allowed maximum angle')

        duty_cycle = self.__map_angle_to_duty_cycle(angle, min_duty_cycle, 
                                                   max_duty_cycle, min_angle, max_angle)
        self.start(duty_cycle)
        time.sleep(wait)
        self.stop()



    def __map_angle_to_duty_cycle(self, angle, min_duty, max_duty, min_angle, max_angle):
        angle = max(min_angle, min(max_angle, angle))
        mapped_duty_cycle = ((angle - min_angle) / (max_angle - min_angle)) * (max_duty - min_duty) + min_duty
        return mapped_duty_cycle
    


    def cleanup(self):
        """
        Perform cleanup
        """
        self.rotate(0)
        


    def __repr__(self):
        min_angle = self.__min_angle_duty_cycle[0]
        max_angle = self.__max_angle_duty_cycle[0]
        return f"Servo Motor <pin={self.pin}, angle={self.__angle}, min={min_angle}, max={max_angle}>"
    


from ..base import ArduinoDevice
from raspidevkit.constants import OUTPUT
from typing import Union
import time


class Relay(ArduinoDevice):
    def __init__(self, arduino, pin: int, commands: dict[str, Union[str, int]]) -> None:
        """
        Create a LED object attached to Arduino

        :param arduino: Arduino master
        :param pin: Pin to attach LED in Arduino
        :param commands: Dictionary of commands with methods as keys.
        Required command configuration:
        ```python
        commands = {
            "turn_on": 1,
            "turn_off": 2,
            "blink": 3
        }
        ```
        """
        pin_setup = {
            str(pin): OUTPUT
        }
        super().__init__(arduino=arduino,
                         pin_setup=pin_setup, 
                         device_type=OUTPUT, 
                         commands=commands)
        all_methods = [
            'turn_on',
            'turn_off'
        ]
        self.validate_commands(all_methods)
        self._method_code = self._map_method_code()
        self._state = False
        self._code_mapping['methods'] = {
            'turn_on': f'digitalWrite({pin}, HIGH);',
            'turn_off': f'digitalWrite({pin}, LOW);'
        }
    


    @property
    def state(self):
        """
        Device state
        """
        return self._state
    


    def turn_on(self):
        """
        Turn on LED. Has no effect if already turned on.
        """
        if not self._state:
            command = self._commands.get('turn_on')
            self._arduino.send_command(command)
            self._state = True



    def turn_off(self):
        """
        Turn off LED. Has no effect if already turned off.
        """
        if self._state:
            command = self._commands.get('turn_off')
            self._arduino.send_command(command)
            self._state = False
        
    

    def _map_method_code(self):
        """
        Create method code mapping
        """
        method_code = {
            'turn_on': f'digitalWrite({self.pin}, HIGH);',
            'turn_off': f'digitalWrite({self.pin}, LOW);',
        }
        return method_code
    


    def cleanup(self):
        """
        Perform cleanup
        """
        self.turn_off()



    def __str__(self):
        """
        Device name
        """
        return "arduino_led"
    


    def __repr__(self):
        return f"Arduino LED <pin={self.pin}, state={self._state}>"
    


import sys
sys.path.append('../')

from ..base import GpioDevice
from raspidevkit.constants import INPUT, OUTPUT
import time
    

class LightSensor(GpioDevice):
    def __init__(self, machine, pin: int):
        """
        Create a light sensor object

        :param machine: Machine instance as this device parent
        :param pin: Pin this device is attached to
        """
        pin_setup = {
            str(pin): INPUT
        }
        super().__init__(machine, pin_setup, device_type=INPUT)



    def read(self) -> int:
        """
        Read the current light value

        :return: Value
        """
        value = 0
        self.gpio_setup(self.pin, OUTPUT)
        self.gpio_write(self.pin, False)
        time.sleep(0.1)
        self.gpio_setup(self.pin, INPUT)
        while not self.gpio_read(self.pin):
            value += 1
        return value



    def __repr__(self):
        return f"Light Sensor <pin={self.pin}, value={self.read()}>"
    


import setuptools

setuptools.setup(
    name="raspidevkit",
    description="Easily control devices with Raspberry Pi",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    version="0.0.5",
    url="https://github.com/raspidevkit/raspidevkit",
    author="DailyLollipops",
    author_email="clarencemadrigal84@gmail.com",
    packages=setuptools.find_namespace_packages(),
    include_package_data=True,
    install_requires=[
        'fake-rpi', 
        'pyserial', 
        'setuptools', 
        'smbus2',
        'wheel',
    ],
    package_data={
        '': ['templates/*', 'templates/.clang-format'],
        'raspidevkit': ['templates/*', 'templates/.clang-format']
    },
)



from .gpio.button import Button
from .gpio.dht_sensor import DHT11, DHT22
from .gpio.hall_effect_sensor import HallEffectSensor
from .gpio.led import Led
from .gpio.rgb_led import RgbLed
from .gpio.buzzer import ActiveBuzzer
from .gpio.buzzer import PassiveBuzzer
from .gpio.light_sensor import LightSensor
from .gpio.pir_motion_sensor import PIRMotionSensor
from .gpio.servo_motor import ServoMotor
from .gpio.relay import Relay
from .gpio.ultrasonic_sensor import UltrasonicSensor

from .drivers.dc_motor.l298n import L298NDriver
from .drivers.dc_motor.l293d import L293DDriver

from .serial.sim808 import Sim808
from .serial.arduino import Arduino



from .__logger import MachineLogger
from .utils import fileutil
from .constants import INPUT, PULL_UP, PULL_DOWN, OUTPUT
from typing import Union
import json
import re
import subprocess
import sys
import time

try:
    import RPi.GPIO as GPIO
except (RuntimeError, ModuleNotFoundError):
    import fake_rpi
    sys.modules['RPi'] = fake_rpi.RPi
    sys.modules['RPi.GPIO'] = fake_rpi.RPi.GPIO
    import RPi.GPIO as GPIO


class GpioInterface:
    def __init__(self, mode: str, logger: MachineLogger) -> None:
        """
        Initialize Machine GPIO Interface

        :param mode: GPIO board setup
        :param logger: Machine Logger object
        """
        self.__logger = logger
        if mode.upper() == 'BCM':
            GPIO.setmode(GPIO.BCM)
        elif mode.upper() == 'BOARD':
            GPIO.setmode(GPIO.BCM)
        else:
            raise ValueError('GPIO mode can only be BCM or BOARD.')
        


    def setup(self, pin: int,  setup: str):
        """
        Explicitly setup a GPIO pin

        :param pin: Pin to setup
        :param setup: Pin mode (INPUT or OUTPUT)
        """
        self.__logger.debug(f'[GPIO][SETUP] {pin}:{setup}')
        if setup.upper() == INPUT:
            GPIO.setup(pin, GPIO.IN)
        elif setup.upper() == PULL_UP:
            GPIO.setup(int(pin), GPIO.IN, pull_up_down=GPIO.PUD_UP)
        elif setup.upper() == PULL_DOWN:
            GPIO.setup(int(pin), GPIO.IN, pull_up_down=GPIO.PUD_DOWN)
        elif setup.upper() == OUTPUT:
            GPIO.setup(pin, GPIO.OUT)
        else:
            raise ValueError('Invalid pin mode')
        


    def write(self, pin: int, value: bool):
        """
        Perform a GPIO output on a pin

        :param pin: The pin to write
        :param value: Can be HIGH or LOW
        """
        self.__logger.debug(f'[GPIO][WRITE] Write on {pin}: {value}')
        if value:
            GPIO.output(pin, GPIO.HIGH)
        else:
            GPIO.output(pin, GPIO.LOW)



    def read(self, pin: int) -> bool:
        """
        Read the value from a GPIO pin
        
        :param pin: The pin to read
        :return: Value, either True of False
        """
        self.__logger.debug(f'[GPIO][READ] Read on: {pin}')
        value = GPIO.input(pin)
        self.__logger.debug(f'[GPIO][READ] Value from {pin}: {value}')
        return value
    


    def cleanup(self):
        """
        Perform cleanup
        """
        GPIO.cleanup()


class I2cInterface:
    def __init__(self, i2cbus: int, logger: MachineLogger) -> None:
        """
        Initialize Machine GPIO Interface

        :param i2cbus: I2C bus to use
        :param logger: Machine Logger object
        """
        self.__logger = logger
        self.__bus = None
        self.__enabled = False
        if i2cbus is not None:
            import smbus2
            self.__bus = smbus2.SMBus(i2cbus)
            self.__enabled = True



    @property
    def enabled(self) -> bool:
        """
        I2C Interface enabled
        """
        return self.__enabled



    def read_byte(self, address: int, force: bool = False) -> int:
        """
        Read a single byte from an I2C address.

        :param address: Device I2C address
        :param force: Force read flag
        :return: Read byte value
        """
        if not self.__enabled:
            raise Exception('I2C bus is not enabled.')
        self.__logger.debug(f'[I2C][READ] Read on {address}')
        value = self.__bus.read_byte(address, force)
        self.__logger.debug(f'[I2C][READ] Returned: {value}')
        return value
    


    def read_word_data(self, address: int, register: int, 
                           force: bool = False) -> int:
        """
        Read a single word (2 bytes) from a given register

        :param address: Device I2C address
        :param register: Register address
        :param force: Force read flag
        :return: 2-byte word
        """
        if not self.__enabled:
            raise Exception('I2C bus is not enabled.')
        self.__logger.debug(f'[I2C][READ] Read word data on {address}-{register}')
        value = self.__bus.read_word_data(address, register, force)
        self.__logger.debug(f'[I2C][READ] Returned: {value}')
        return value
    


    def read_byte_data(self, address: int, register: int, 
                           force: bool = False) -> int:
        """
        Read a single byte from a designated register
        
        :param address: Device I2C address
        :param register: Register address
        :param force: Force read flag
        :return: Read byte value
        """
        if not self.__enabled:
            raise Exception('I2C bus is not enabled.')
        self.__logger.debug(f'[I2C][READ] Read byte data on {address}-{register}')
        value = self.__bus.read_byte_data(address, register, force)
        self.__logger.debug(f'[I2C][READ] Returned: {value}')
        return value
    


    def read_block_data(self, address: int, register: int, 
                            length: int = 32, force: bool = False) -> list[int]:
        """
        Read a block of data from a given register

        :param address: Device I2C address
        :param register: Register address
        :param length: Block length, default to 32 bytes
        :param force: Force read flag
        :return: List of bytes
        """
        if not self.__enabled:
            raise Exception('I2C bus is not enabled.')
        self.__logger.debug(f'[I2C][READ] Read block data ({length}) on {address}-{register}')
        value = self.__bus.read_i2c_block_data(address, register, length, force)
        self.__logger.debug(f'[I2C][READ] Returned: {value}')
        return value
    


    def write_byte(self, address: int, value: int, force: bool = False):
        """
        Write a single byte to an I2C address

        :param address: Device I2C address
        :param value: Value to write
        :param force: Force write flag
        """
        if not self.__enabled:
            raise Exception('I2C bus is not enabled.')
        self.__logger.debug(f'[I2C][WRITE] Write on {address}: {value}')
        self.__bus.write_byte(address, value, force)



    def write_word_data(self, address: int, register: int, 
                            value:int, force: bool = False):
        """
        Write a single word (2 bytes) to a given register

        :param address: Device I2C address
        :param register: Register address
        :param value: Word value to write
        :param force: Force write flag
        """
        if not self.__enabled:
            raise Exception('I2C bus is not enabled.')
        self.__logger.debug(f'[I2C][WRITE] Write word data on {address}-{register}: {value}')
        self.__bus.write_word_data(address, register, value, force)


        
    def write_byte_data(self, address: int, register: int, 
                            value: int, force: bool = False):
        """
        Write a byte to a given register

        :param address: Device I2C address
        :param register: Register address
        :param value: Byte value to write
        :param force: Force write flag
        """
        if not self.__enabled:
            raise Exception('I2C bus is not enabled.')
        self.__logger.debug(f'[I2C][WRITE] Write yte data on {address}-{register}: {value}')
        self.__bus.write_byte_data(address, register, value, force)


    
    def write_block_data(self, address: int, register: int, 
                             data: list[int], force: bool = False):
        """
        Write a block of byte data to a given register

        :param address: Device I2C address
        :param register: Register address
        :param data: List of byte values to write
        :param force: Force write flag
        """
        if not self.__enabled:
            raise Exception('I2C bus is not enabled.')
        self.__logger.debug(f'[I2C][WRITE] Write block data on {address}-{register}: {data}')
        self.__bus.write_block_data(address, register, data, force)



class ArduinoCliInterface:
    def __init__(self, arduino_cli_path: Union[str, None], logger: MachineLogger) -> None:
        """
        Initialize Arduino CLI interface

        :param arduino_cli_path: Path to Arduino CLI
        :param logger: Machine Logger object
        """
        self.__logger = logger
        if arduino_cli_path is None:
            self.__path = None
            self.__enabled = None
            self.__formatting = False
        else:
            self.__enabled = self.__is_arduino_cli_installed(arduino_cli_path)
            self.__path = arduino_cli_path if self.__enabled else None
            self.__formatting = self.__is_clang_format_installed()
            self.__setup_libraries()



    @property
    def path(self) -> Union[str, None]:
        """
        Arduino CLI path
        """
        return self.__path
    


    @property
    def enabled(self) -> bool:
        """
        Arduino CLI enabled and working
        """
        return self.__enabled
    


    @property
    def formatting(self) -> bool:
        """
        Formatting enabled. Useful if manually generating code
        """
        return self.__formatting
    


    def __is_arduino_cli_installed(self, path: str) -> bool:
        """
        Check if arduino is found in the path given

        :param arduino_cli_path: Arduino CLI path
        :return: Arduino CLI installed
        """
        try:
            subprocess.run(['arduino-cli', 'version'],
                           stdout=subprocess.PIPE, 
                           stderr=subprocess.PIPE, 
                           check=True,
                           cwd=path,
                           shell=True)
            self.__logger.info('Arduino CLI found.')
            return True
        except subprocess.CalledProcessError:
            self.__logger.warning('Arduino CLI not found.')
            return False
        


    def __is_clang_format_installed(self) -> bool:
        """
        Check if clang-format is installed
        """
        try:
            subprocess.run(['clang-format', '--version'], 
                           stdout=subprocess.PIPE, 
                           stderr=subprocess.PIPE, 
                           check=True, 
                           shell=True)
            self.__logger.info('Clang-format found.')
            return True
        except subprocess.CalledProcessError:
            self.__logger.warning('Clang-format not found. Arduino generated code will not be formatted')
            return False
        


    def __setup_libraries(self):
        """
        Check for required libraries and install as neccessary
        """
        if not self.__enabled:
            self.__logger.info(f'Skipping arduino library setup')
            return
        
        required_libraries = {
            'Servo': '1.2.1',
        }
        installed_libraries = self.get_installed_libraries()
        for name, version in required_libraries.items():
            if name not in installed_libraries.keys():
                self.__logger.info(f'Installing {name}@{version}', to_stdout=True)
                self.install_library(name=name, version=version)
            elif installed_libraries.get(name) != version:
                self.__logger.info(f'Updating {name} to {version}', to_stdout=True)
                self.install_library(name=name, version=version)
            else:
                # Skip
                pass



    def get_boards(self) -> dict:
        """
        Get available Arduino boards
        """
        if not self.__enabled:
            raise Exception('Arduino CLI not found')
        
        result = subprocess.run(['arduino-cli', 'board', 'listall',
                                 '--format', 'json'], 
                                check=True, 
                                cwd=self.__path, 
                                shell=True, 
                                capture_output=True, 
                                text = True)
        result = json.loads(result.stdout)
        boards = result['boards']

        board_mapping = {}
        for board in boards:
            name = board['name']
            fqbn = board['fqbn']
            board_mapping[name] = fqbn
        
        return board_mapping
    


    def get_installed_libraries(self) -> dict:
        """
        Get a list of installed libraries in the arduino

        :return: List of installed libraries and version
        """
        if not self.__enabled:
            raise Exception('Arduino CLI not found')
        
        result = subprocess.run(['arduino-cli', 'lib', 'list',
                                 '--format', 'json'], 
                                check=True, 
                                cwd=self.__path, 
                                shell=True, 
                                capture_output=True, 
                                text = True)
        libraries = json.loads(result.stdout)

        library_info_map = {}
        for library in libraries:
            name = library['library']['name']
            version = library['library']['version']
            library_info_map[name] = version

        return library_info_map



    def install_library(self, name: str, version: str = 'latest',
                                git_url: str = '', zip_path: str = ''):
        """
        Install an Arduino library
        Priority: zip_path > git_url > name

        :param name: Arduino library name
        :param version: Library version
        :param git_url: Git url of library
        :param zip_path: Path to library zip
        """
        version_pattern = r'(\d+\.\d+\.\d+)'
        if version != 'latest' and not re.search(version_pattern, version):
            raise Exception('Incorrect version format')
        
        command = ['arduino-cli', 'lib', 'install']
        if zip_path:
            command.append('--zip-path')
            command.append(zip_path)
            self.__logger.info(f'Installing library via zip path: {zip_path}')
        elif git_url:
            command.append('--git_url')
            url = f'{git_url}{'#' + version if version != 'latest' else ''}'
            command.append(url)
            self.__logger.info(f'Installing library via git url: {git_url}')
        else:
            name = f'{name}{'@' + version if version != 'latest' else ''}'
            command.append(name)
            self.__logger.info(f'Installing library via name: {name}')
        
        result = subprocess.run(command, 
                                cwd=self.__path, 
                                shell=True, 
                                capture_output=True, 
                                text = True)
        
        if result.returncode == 1:
            raise Exception(result.stderr)
            


    def uninstall_library(self, library: str):
        """
        Uninstall an Arduino library

        :param library: Library name
        """
        if library not in self.get_installed_libraries():
            raise Exception('Library not installed')
        
        result = subprocess.run(['arduino-cli', 'lib',
                                 'uninstall', library], 
                                cwd=self.__path, 
                                shell=True, 
                                capture_output=True, 
                                text = True)
        
        if result.returncode == 1:
            raise Exception(result.stderr)
        

    def compile_code(self, file_path: str, fqbn: str):
        """
        Compile arduino code

        :param file_path: Sketch file path (should be a directory)
        :param fqbn: Fully Qualified Board Name
        :raises Exception: If build error occurs
        """
        if not self.__enabled:
            raise Exception('Arduino CLI not found')

        abs_path = fileutil.get_absolute_path(file_path)
        result = subprocess.run(['arduino-cli', 'compile',
                                 '--fqbn', fqbn, abs_path], 
                                cwd=self.__path, 
                                shell=True, 
                                capture_output=True, 
                                text = True)
        
        if result.returncode == 1:
            raise Exception(result.stderr)
        


    def upload_code(self, file_path: str, port: str, 
                            fqbn: str, compile: bool = True):
        """
        Upload a sketch to the arduino

        :param file_path: Sketch file path (should be a directory)
        :param port: Arduino port
        :param fqbn: Fully Qualified Board Name
        :param compile: Try to compile the arduino first before upload
        """
        if not self.__enabled:
            raise Exception('Arduino CLI not found')
        
        self.__logger.info('Starting arduino code upload')
        start = time.time()
        abs_path = fileutil.get_absolute_path(file_path)
        self.compile_code(abs_path, fqbn)
        result = subprocess.run(['arduino-cli', 'upload',
                                 '-p', port, '--fqbn', fqbn, 
                                 abs_path], 
                                cwd=self.__path, 
                                shell=True, 
                                capture_output=True, 
                                text = True)
        end = time.time()
        if result.returncode == 1:
            raise Exception(result.stderr)
        self.__logger.info(f'Arduino code uploaded in {end - start} seconds')
